#/********************************************************************
# Cayenue/cayenue/panels/settings/alarm.py 
#
# Copyright (c) 2024  Stephen Rhodes
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#*********************************************************************/

import os
from PyQt6.QtWidgets import QSpinBox, QGridLayout, QWidget, \
    QLabel, QComboBox, QSlider, QCheckBox, QLabel
from PyQt6.QtCore import Qt

class AlarmOptions(QWidget):
    def __init__(self, mw):
        super().__init__()
        self.mw = mw

        self.bufferSizeKey = "settings/bufferSize"
        self.lagTimeKey = "settings/lagTime"
        self.alarmSoundFileKey = "settings/alarmSoundFile"
        self.alarmSoundVolumeKey = "settings/alarmSoundVolume"
        self.showDisplayAlarmKey = "settings/showDisplayAlarm"
        self.savePictureKey = "settings/savePictureOnAlarm"

        self.spnBufferSize = QSpinBox()
        self.spnBufferSize.setMinimum(1)
        self.spnBufferSize.setMaximum(60)
        self.spnBufferSize.setMaximumWidth(80)
        self.spnBufferSize.setValue(int(self.mw.settings.value(self.bufferSizeKey, 10)))
        self.spnBufferSize.valueChanged.connect(self.spnBufferSizeChanged)
        lblBufferSize = QLabel("Pre-Alarm Buffer Size (in seconds)")

        self.spnLagTime = QSpinBox()
        self.spnLagTime.setMinimum(1)
        self.spnLagTime.setMaximum(60)
        self.spnLagTime.setMaximumWidth(80)
        self.spnLagTime.setValue(int(self.mw.settings.value(self.lagTimeKey, 5)))
        self.spnLagTime.valueChanged.connect(self.spnLagTimeChanged)
        lblLagTime = QLabel("Post-Alarm Lag Time (in seconds)")

        self.cmbSoundFiles = QComboBox()
        d = f'{self.mw.getLocation()}/cayenue/resources'
        sounds = [f for f in os.listdir(d) if os.path.isfile(os.path.join(d, f)) and f.endswith(".mp3")]
        self.cmbSoundFiles.addItems(sounds)
        self.cmbSoundFiles.currentTextChanged.connect(self.cmbSoundFilesChanged)
        self.cmbSoundFiles.setCurrentText(self.mw.settings.value(self.alarmSoundFileKey, "drops.mp3"))
        lblSoundFiles = QLabel("Alarm Sounds")
        self.sldAlarmVolume = QSlider(Qt.Orientation.Horizontal)
        self.sldAlarmVolume.setValue(int(self.mw.settings.value(self.alarmSoundVolumeKey, 80)))
        self.sldAlarmVolume.valueChanged.connect(self.sldAlarmVolumeChanged)

        self.chkShowDisplay = QCheckBox("Show Alarms on Display")
        self.chkShowDisplay.setChecked(int(self.mw.settings.value(self.showDisplayAlarmKey, 1)))
        self.chkShowDisplay.stateChanged.connect(self.chkShowDisplayClicked)

        self.chkSavePicture = QCheckBox("Save Picture for Alarms")
        self.chkSavePicture.setChecked(int(self.mw.settings.value(self.savePictureKey, 1)))
        self.chkSavePicture.stateChanged.connect(self.chkSavePictureClicked)

        pnlSoundFile = QWidget()
        lytSoundFile =  QGridLayout(pnlSoundFile)
        lytSoundFile.addWidget(lblSoundFiles,        0, 0, 1, 1)
        lytSoundFile.addWidget(self.cmbSoundFiles,   0, 1, 1, 1)
        lytSoundFile.addWidget(self.sldAlarmVolume,  0, 2, 1, 1)
        lytSoundFile.addWidget(QLabel(),             1, 0, 1, 3)
        lytSoundFile.addWidget(self.chkShowDisplay,  2, 0, 1, 2)
        lytSoundFile.addWidget(self.chkSavePicture,  3, 0, 1, 2)
        lytSoundFile.setColumnStretch(1, 10)

        pnlBuffer = QWidget()
        lytBuffer = QGridLayout(pnlBuffer)
        lytBuffer.addWidget(lblBufferSize,          1, 0, 1, 3)
        lytBuffer.addWidget(self.spnBufferSize,     1, 3, 1, 1)
        lytBuffer.addWidget(lblLagTime,             2, 0, 1, 3)
        lytBuffer.addWidget(self.spnLagTime,        2, 3, 1, 1)
        lytBuffer.addWidget(pnlSoundFile,           3, 0, 1, 4)
        lytBuffer.setContentsMargins(0, 0, 0, 0)

        lytMain = QGridLayout(self)
        lytMain.addWidget(pnlBuffer,  0, 0, 1, 1)
        lytMain.addWidget(QLabel(),   1, 0, 1, 1)
        lytMain.setRowStretch(1, 10)

    def spnBufferSizeChanged(self, i):
        self.mw.settings.setValue(self.bufferSizeKey, i)

    def spnLagTimeChanged(self, i):
        self.mw.settings.setValue(self.lagTimeKey, i)

    def cmbSoundFilesChanged(self, value):
        self.mw.settings.setValue(self.alarmSoundFileKey, value)

    def sldAlarmVolumeChanged(self, value):
        self.mw.settings.setValue(self.alarmSoundVolumeKey, value)

    def chkShowDisplayClicked(self, state):
        self.mw.settings.setValue(self.showDisplayAlarmKey, state)

    def chkSavePictureClicked(self, state):
        self.mw.settings.setValue(self.savePictureKey, state)


#/********************************************************************
# Cayenue/cayenue/panels/cameras/networktab.py 
#
# Copyright (c) 2023  Stephen Rhodes
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#*********************************************************************/

from PyQt6.QtWidgets import QCheckBox, QLineEdit, QGridLayout, QWidget, QLabel
from PyQt6.QtCore import QRegularExpression
from PyQt6.QtGui import QRegularExpressionValidator

class NetworkTab(QWidget):
    def __init__(self, cp):
        super().__init__()
        self.cp = cp

        ipRange = "(?:[0-1]?[0-9]?[0-9]|2[0-4][0-9]|25[0-5])" 
        ipRegex = QRegularExpression("^" + ipRange + "\\." + ipRange + "\\." + ipRange + "\\." + ipRange + "$")
        ipValidator = QRegularExpressionValidator(ipRegex, self)           

        self.chkDHCP = QCheckBox("DHCP enabled")
        self.chkDHCP.clicked.connect(cp.onEdit)
        self.chkDHCP.clicked.connect(self.onChkDHCPChecked)
        self.txtIPAddress = QLineEdit()
        self.txtIPAddress.setValidator(ipValidator)
        self.txtIPAddress.textEdited.connect(cp.onEdit)
        lblIPAddress = QLabel("Address")
        self.txtSubnetMask = QLineEdit()
        self.txtSubnetMask.setValidator(ipValidator)
        self.txtSubnetMask.textEdited.connect(cp.onEdit)
        lblSubnetMask = QLabel("Subnet")
        self.txtDefaultGateway = QLineEdit()
        self.txtDefaultGateway.setValidator(ipValidator)
        self.txtDefaultGateway.textEdited.connect(cp.onEdit)
        lblDefaultGateway = QLabel("Gateway")
        self.txtDNS = QLineEdit()
        self.txtDNS.setValidator(ipValidator)
        self.txtDNS.textEdited.connect(cp.onEdit)
        lblDNS = QLabel("DNS")

        lytMain = QGridLayout(self)
        lytMain.addWidget(self.chkDHCP,           0, 1, 1, 1)
        lytMain.addWidget(lblIPAddress,           1, 0, 1, 1)
        lytMain.addWidget(self.txtIPAddress,      1, 1, 1, 1)
        lytMain.addWidget(lblSubnetMask,          2, 0, 1, 1)
        lytMain.addWidget(self.txtSubnetMask,     2, 1, 1, 1)
        lytMain.addWidget(lblDefaultGateway,      3, 0, 1, 1)
        lytMain.addWidget(self.txtDefaultGateway, 3, 1, 1, 1)
        lytMain.addWidget(lblDNS,                 4, 0, 1, 1)
        lytMain.addWidget(self.txtDNS,            4, 1, 1, 1)

    def fill(self, onvif_data):
        self.chkDHCP.setChecked(onvif_data.dhcp_enabled())
        self.txtIPAddress.setText(onvif_data.ip_address_buf())
        self.txtDefaultGateway.setText(onvif_data.default_gateway_buf())
        self.txtDNS.setText(onvif_data.dns_buf())
        self.txtSubnetMask.setText(onvif_data.mask_buf())
        self.txtIPAddress.setEnabled(not onvif_data.dhcp_enabled())
        self.txtDefaultGateway.setEnabled(not onvif_data.dhcp_enabled())
        self.txtSubnetMask.setEnabled(not onvif_data.dhcp_enabled())
        self.txtDNS.setEnabled(not onvif_data.dhcp_enabled())
        self.setEnabled(len(onvif_data.ip_address_buf()))
        self.cp.onEdit()

    def edited(self, onvif_data):
        result = False
        if self.isEnabled():
            if not onvif_data.dhcp_enabled() == self.chkDHCP.isChecked():
                result = True
            if not onvif_data.ip_address_buf() == self.txtIPAddress.text():
                result = True
            if not onvif_data.default_gateway_buf() == self.txtDefaultGateway.text():
                result = True
            if not onvif_data.dns_buf() == self.txtDNS.text():
                result = True
            if not onvif_data.mask_buf() == self.txtSubnetMask.text():
                result = True

        return result
    
    def update(self, onvif_data):
        if self.edited(onvif_data):
            onvif_data.setDHCPEnabled(self.chkDHCP.isChecked())
            onvif_data.setIPAddressBuf(self.txtIPAddress.text())
            onvif_data.setDefaultGatewayBuf(self.txtDefaultGateway.text())
            onvif_data.setDNSBuf(self.txtDNS.text())
            onvif_data.setMaskBuf(self.txtSubnetMask.text())
            onvif_data.startUpdateNetwork()

    def onChkDHCPChecked(self):
        checked = self.chkDHCP.isChecked()
        self.txtIPAddress.setEnabled(False)
        self.txtDefaultGateway.setEnabled(False)
        self.txtSubnetMask.setEnabled(False)
        self.txtDNS.setEnabled(False)



#/********************************************************************
# Cayenue/cayenue/panels/cameras/ptztab.py 
#
# Copyright (c) 2023  Stephen Rhodes
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#*********************************************************************/

from PyQt6.QtWidgets import QPushButton, QGridLayout, QWidget, QCheckBox
from cayenue.enums import ProxyType

class PTZTab(QWidget):
    def __init__(self, cp):
        super().__init__()
        self.cp = cp

        self.btn1 = QPushButton("1")
        self.btn1.setMaximumWidth(52)
        self.btn1.pressed.connect(lambda val=1: self.presetButtonClicked(val))
        self.btn2 = QPushButton("2")
        self.btn2.setMaximumWidth(52)
        self.btn2.pressed.connect(lambda val=2: self.presetButtonClicked(val))
        self.btn3 = QPushButton("3")
        self.btn3.setMaximumWidth(52)
        self.btn3.pressed.connect(lambda val=3: self.presetButtonClicked(val))
        self.btn4 = QPushButton("4")
        self.btn4.setMaximumWidth(52)
        self.btn4.pressed.connect(lambda val=4: self.presetButtonClicked(val))
        self.btn5 = QPushButton("5")
        self.btn5.setMaximumWidth(52)
        self.btn5.pressed.connect(lambda val=5: self.presetButtonClicked(val))

        self.btnLeft = QPushButton("<")
        self.btnLeft.setMaximumWidth(52)
        self.btnLeft.pressed.connect(   lambda x=-0.5, y=0.0,  z=0.0 : self.move(x, y, z))
        self.btnLeft.released.connect(self.stopPanTilt)
        self.btnRight = QPushButton(">")
        self.btnRight.setMaximumWidth(52)
        self.btnRight.pressed.connect(  lambda x=0.5,  y=0.0,  z=0.0 : self.move(x, y, z))
        self.btnRight.released.connect(self.stopPanTilt)
        self.btnUp = QPushButton("^")
        self.btnUp.setMaximumWidth(52)
        self.btnUp.pressed.connect(     lambda x=0.0,  y=0.5,  z=0.0 : self.move(x, y, z))
        self.btnUp.released.connect(self.stopPanTilt)
        self.btnDown = QPushButton("v")
        self.btnDown.setMaximumWidth(52)
        self.btnDown.pressed.connect(   lambda x=0.0,  y=-0.5, z=0.0 : self.move(x, y, z))
        self.btnDown.released.connect(self.stopPanTilt)
        self.btnZoomIn = QPushButton("+")
        self.btnZoomIn.setMaximumWidth(52)
        self.btnZoomIn.pressed.connect( lambda x=0.0,  y=0.0,  z=0.5 : self.move(x, y, z))
        self.btnZoomIn.released.connect(self.stopZoom)
        self.btnZoomOut = QPushButton("-")
        self.btnZoomOut.setMaximumWidth(52)
        self.btnZoomOut.pressed.connect( lambda x=0.0,  y=0.0, z=-0.5 : self.move(x, y, z))
        self.btnZoomOut.released.connect(self.stopZoom)

        self.chkSet = QCheckBox("Set Preset Position")

        lytMain = QGridLayout(self)
        lytMain.addWidget(self.btn1,   0, 0, 1, 1)
        lytMain.addWidget(self.btn2,   1, 0, 1, 1)
        lytMain.addWidget(self.btn3,   2, 0, 1, 1)
        lytMain.addWidget(self.btn4,   3, 0, 1, 1)
        lytMain.addWidget(self.btn5,   4, 0, 1, 1)

        lytMain.addWidget(self.btnLeft,    1, 2, 1, 1)
        lytMain.addWidget(self.btnUp,      0, 3, 1, 1)
        lytMain.addWidget(self.btnDown,    2, 3, 1, 1)
        lytMain.addWidget(self.btnRight,   1, 4, 1, 1)

        lytMain.addWidget(self.btnZoomIn,  3, 4, 1, 1)
        lytMain.addWidget(self.btnZoomOut, 4, 4, 1, 1)

        lytMain.addWidget(self.chkSet,     4, 1, 1, 3)

    def presetButtonClicked(self, n):
        camera = self.cp.getCurrentCamera()
        if camera:
            if self.chkSet.isChecked():
                camera.onvif_data.preset = n
                if self.cp.mw.settingsPanel.proxy.proxyType == ProxyType.CLIENT:
                    arg = "SET PRESET\n\n" + camera.onvif_data.toJSON() + "\r\n"
                    self.cp.mw.client.transmit(bytearray(arg, 'utf-8'))
                else:
                    camera.onvif_data.startSetGotoPreset()
            else:
                camera.onvif_data.preset = n
                if self.cp.mw.settingsPanel.proxy.proxyType == ProxyType.CLIENT:
                    arg = "GOTO PRESET\n\n" + camera.onvif_data.toJSON() + "\r\n"
                    self.cp.mw.client.transmit(bytearray(arg, 'utf-8'))
                else:
                    camera.onvif_data.startSet()

    def move(self, x, y, z):
        camera = self.cp.getCurrentCamera()
        if camera:
            camera.onvif_data.x = x
            camera.onvif_data.y = y
            camera.onvif_data.z = z
            if self.cp.mw.settingsPanel.proxy.proxyType == ProxyType.CLIENT:
                arg = "MOVE\n\n" + camera.onvif_data.toJSON() + "\r\n"
                self.cp.mw.client.transmit(bytearray(arg, 'utf-8'))
            else:
                camera.onvif_data.startMove()

    def stopPanTilt(self):
        camera = self.cp.getCurrentCamera()
        if camera:
            camera.onvif_data.stop_type = 0
            if self.cp.mw.settingsPanel.proxy.proxyType == ProxyType.CLIENT:
                arg = "STOP\n\n" + camera.onvif_data.toJSON() + "\r\n"
                self.cp.mw.client.transmit(bytearray(arg, 'utf-8'))
            else:
                camera.onvif_data.startStop()

    def stopZoom(self):
        camera = self.cp.getCurrentCamera()
        if camera:
            camera.onvif_data.stop_type = 1
            if self.cp.mw.settingsPanel.proxy.proxyType == ProxyType.CLIENT:
                arg = "STOP\n\n" + camera.onvif_data.toJSON() + "\r\n"
                self.cp.mw.client.transmit(bytearray(arg, 'utf-8'))
            else:
                camera.onvif_data.startStop()

    def fill(self, onvif_data):
        self.setEnabled(True)
        self.chkSet.setChecked(False)


#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# Copyright (c) Megvii Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#******************************************************************************/

import torch.nn as nn

from .yolo_head import YOLOXHead
from .yolo_pafpn import YOLOPAFPN


class YOLOX(nn.Module):
    """
    YOLOX model module. The module list is defined by create_yolov3_modules function.
    The network returns loss values from three YOLO layers during training
    and detection results during test.
    """

    def __init__(self, backbone=None, head=None):
        super().__init__()
        if backbone is None:
            backbone = YOLOPAFPN()
        if head is None:
            head = YOLOXHead(80)

        self.backbone = backbone
        self.head = head

    def forward(self, x, targets=None):
        # fpn output content features of [dark3, dark4, dark5]
        fpn_outs = self.backbone(x)

        if self.training:
            assert targets is not None
            loss, iou_loss, conf_loss, cls_loss, l1_loss, num_fg = self.head(
                fpn_outs, targets, x
            )
            outputs = {
                "total_loss": loss,
                "iou_loss": iou_loss,
                "l1_loss": l1_loss,
                "conf_loss": conf_loss,
                "cls_loss": cls_loss,
                "num_fg": num_fg,
            }
        else:
            outputs = self.head(fpn_outs)

        return outputs



#/********************************************************************
# Cayenue/cayenue/manager.py 
#
# Copyright (c) 2024  Stephen Rhodes
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#*********************************************************************/

from time import sleep
from loguru import logger
from PyQt6.QtCore import QRectF, QSize, Qt, QSizeF, QPointF

class Manager():
    def __init__(self, mw):
        self.players = {}
        self.ordinals = {}
        self.sizes = {}
        self.thread_lock = False
        self.mw = mw

    def lock(self):
        # the lock protects the player list
        while self.thread_lock:
            sleep(0.001)
        self.thread_lock = True

    def unlock(self):
        self.thread_lock = False

    def startPlayer(self, player):
        self.lock()
        if not player.uri in self.ordinals.keys():
            ordinal = self.nextOrdinal()
    
            if player.isCameraStream():
                if camera := self.mw.cameraPanel.getCamera(player.uri):
                    if camera.companionURI(player.uri) in self.ordinals.keys():
                        ordinal = self.ordinals[camera.companionURI(player.uri)]
                    camera.setOrdinal(ordinal)
            
            self.ordinals[player.uri] = ordinal

        self.players[player.uri] = player
        player.start()
        self.unlock()

    def nextOrdinal(self):
        values = set(self.ordinals.values())
        size = len(values)
        for i in range(size):
            if not i in values:
                return i
        return size
        
    def getPlayer(self, uri):
        if not uri: return None
        return self.players.get(uri)
    
    def getCurrentPlayer(self):
        return self.getPlayer(self.mw.glWidget.focused_uri)
    
    def countPlayers(self):
        return len(self.players)
    
    def waitForAllPlayersClosed(self):
        count = 0
        while self.countPlayers():
            sleep(0.1)
            count += 1
            if count > 50:
                raise RuntimeError("Not all players closed in the alloted time")
    
    def hasFilePlayers(self):
        result = False
        self.lock()
        for player in self.players.values():
            if not player.isCameraStream():
                result = True
                break
        self.unlock()
        return result
    
    def getStreamPairProfiles(self, uri):
        result = []
        if uri:
            if camera := self.mw.cameraPanel.getCamera(uri):
                if displayProfile := camera.getDisplayProfile():
                    result.append(displayProfile)
                if camera.displayProfileIndex() != camera.recordProfileIndex():
                    if recordProfile := camera.getRecordProfile():
                        result.append(recordProfile)
        return result
    
    def getStreamPairURIs(self, uri):
        result = []
        profiles = self.getStreamPairProfiles(uri)
        for profile in profiles:
            result.append(profile.uri())
        return result
    
    def getStreamPairPlayers(self, uri):
        result = []
        profiles = self.getStreamPairProfiles(uri)
        for profile in profiles:
            if player := self.getPlayer(profile.uri()):
                result.append(player)
        return result
    
    def getStreamPairTimers(self, uri):
        result = []
        profiles = self.getStreamPairProfiles(uri)
        for profile in profiles:
            if timer := self.mw.timers.get(profile.uri(), None):
                result.append(timer)
        return result
    
    def removeKeys(self, uri):
        if not uri: return
        if uri in self.ordinals:
            del self.ordinals[uri]
        if uri in self.sizes:
            del self.sizes[uri]
    
    def recording(self):
        self.lock()
        result = False
        for player in self.players.values():
            if player.isRecording():
                result = True
                break
        self.unlock()
        return result

    def removePlayer(self, uri):
        if not uri: return
        self.lock()
        if uri in self.players:
            if not self.players[uri].request_reconnect:
                self.removeKeys(uri)
            self.players[uri].lock()
            del self.players[uri]

        recording = False
        for player in self.players.values():
            if player.isRecording():
                recording = True
                break
        if not recording:
            self.mw.enable_disk_mgmt = False

        if not len(self.players):
            self.mw.glWidget.force_clear = True

        self.unlock()

    def playerShutdownWait(self, uri):
        if not uri: return
        if player := self.getPlayer(uri):
            player.requestShutdown()
            count = 0
            while uri in self.players:
                sleep(0.01)
                count += 1
                if count > 100:
                    logger.error(f'Player did not complete shut down during allocated time interval: {uri}')
                    break

    def getMostCommonAspectRatio(self):
        ratio_counter = {}
        for size in self.sizes.values():
            ratio = int(100 * size.width() / size.height())
            if not ratio in ratio_counter.keys():
                ratio_counter[ratio] = 1
            else:
                ratio_counter[ratio] += 1

        highest_count_key = -1
        if ratio_counter:
            keys = list(ratio_counter.keys())
            if len(keys):
                highest_count_key = keys[0]
                highest_count = ratio_counter[highest_count_key]
            for key in keys:
                if ratio_counter[key] > highest_count:
                    highest_count = ratio_counter[key]
                    highest_count_key = key

        return highest_count_key
    
    def getBlankSpace(self, canvas_size: QSize) -> list[QRectF]:
        # after determining the size of the composite image which is the aggregate containing all streams, 
        # compute the rectangles required to fill out the blank space not occupied by any current stream

        ar = self.getMostCommonAspectRatio()
        if ar == -1:
            ar = 177
            
        aspect_ratio = ar / 100
        blanks = []
        num_rows, num_cols = self.computeRowsCols(canvas_size, aspect_ratio)
        if not num_rows:
            blanks.append(QRectF(QPointF(0, 0), QSizeF(canvas_size)))
            return blanks
        
        composite_size = QSizeF(num_cols * aspect_ratio, num_rows)
        composite_size.scale(QSizeF(canvas_size), Qt.AspectRatioMode.KeepAspectRatio)
        im_w = composite_size.toSize().width()
        im_h = composite_size.toSize().height()
        cv_w = canvas_size.width()
        cv_h = canvas_size.height()
        if im_h == cv_h:
            blank_w = (cv_w - im_w)/2
            blanks.append(QRectF(0, 0, blank_w, im_h))
            blanks.append(QRectF(im_w + blank_w, 0, blank_w, im_h))
        if im_w == cv_w:
            blank_h = (cv_h - im_h)/2
            blanks.append(QRectF(0, 0, im_w, blank_h))
            blanks.append(QRectF(0, im_h + blank_h, im_w, blank_h))
        for i in range(num_rows * num_cols):
            if i not in self.ordinals.values():
                blanks.append(self.rectForOrdinal(i, canvas_size, aspect_ratio, num_rows, num_cols))
        return blanks

    def rectForOrdinal(self, ordinal: int, canvas_size: QSize, aspect_ratio: float, num_rows: int, num_cols: int) -> QRectF:
        if not num_rows or ordinal < 0:
            return QRectF(QPointF(0, 0, QSizeF(canvas_size)))

        col = ordinal % num_cols
        row = int(ordinal / num_cols)

        composite_size = QSizeF(num_cols * aspect_ratio, num_rows)
        composite_size.scale(QSizeF(canvas_size), Qt.AspectRatioMode.KeepAspectRatio)

        cell_width = composite_size.width() / num_cols
        cell_height = composite_size.height() / num_rows

        image_size = QSizeF(aspect_ratio, 1)
        image_size.scale(cell_width, cell_height, Qt.AspectRatioMode.KeepAspectRatio)
        w = image_size.width()
        h = image_size.height()

        x_offset = (canvas_size.width() - composite_size.width() + (cell_width - w)) / 2
        y_offset = (canvas_size.height() - composite_size.height() + (cell_height - h)) / 2
        
        x = (col * cell_width) + x_offset
        y = (row * cell_height) + y_offset

        return QRectF(x, y, w, h)


    def computeRowsCols(self, size_canvas, aspect_ratio):
        num_cells = len(set(self.ordinals.values()))

        valid_layouts = []
        for i in range(1, num_cells+1):
            for j in range(num_cells, 0, -1):
                if ((i * j) >= num_cells):
                    if (((i-1)*j) < num_cells) and ((i*(j-1)) < num_cells):
                        valid_layouts.append(QSize(i, j))

        index = -1
        min_ratio = 0
        first_pass = True
        for i, layout in enumerate(valid_layouts):
            if not layout.width():
                continue
            composite = (aspect_ratio * layout.height()) / layout.width()
            if not composite or not size_canvas.height():
                continue
            ratio = (size_canvas.width() / size_canvas.height()) / composite
            optimize = abs(1 - ratio)
            if first_pass:
                first_pass = False
                min_ratio = optimize
                index = i
            else:
                if optimize < min_ratio:
                    min_ratio = optimize
                    index = i

        if index == -1:
            return 0, 0
        
        return valid_layouts[index].width(), valid_layouts[index].height()

    def displayRect(self, uri, canvas_size):
        ar = self.getMostCommonAspectRatio()
        if ar == -1:
            if player := self.getPlayer(uri):
                if player.desired_aspect:
                    ar = player.desired_aspect
                else:
                    ar = player.width() / player.height()
        
        num_rows, num_cols = self.computeRowsCols(canvas_size, ar / 100)
        if num_cols == 0:
            return QRectF(QPointF(0, 0), QSizeF(canvas_size))

        ordinal = -1
        if uri in self.ordinals.keys():
            ordinal = self.ordinals[uri]
        else:
            return QRectF(QPointF(0, 0), QSizeF(canvas_size))

        if ordinal > num_rows * num_cols - 1:
            ordinal = self.nextOrdinal()
            uris = self.getStreamPairURIs(uri)
            for u in uris:
                if u in self.ordinals.keys():
                    self.ordinals[u] = ordinal
        
        col = ordinal % num_cols
        row = int(ordinal / num_cols)

        composite_size = QSizeF()
        if num_rows:
            composite_size = QSizeF(num_cols * ar / 100, num_rows)
            composite_size.scale(QSizeF(canvas_size), Qt.AspectRatioMode.KeepAspectRatio)

        cell_width = composite_size.width() / num_cols
        cell_height = composite_size.height() / num_rows

        image_size = QSizeF(ar, 100)
        if uri in self.sizes.keys():
            image_size = QSizeF(self.sizes[uri])

        image_size.scale(cell_width, cell_height, Qt.AspectRatioMode.KeepAspectRatio)
        w = image_size.width()
        h = image_size.height()

        x_offset = (canvas_size.width() - composite_size.width() + (cell_width - w)) / 2
        y_offset = (canvas_size.height() - composite_size.height() + (cell_height - h)) / 2
        
        x = (col * cell_width) + x_offset
        y = (row * cell_height) + y_offset

        return QRectF(x, y, w, h)



#/********************************************************************
# Cayenue/cayenue/panels/cameras/imagetab.py 
#
# Copyright (c) 2023  Stephen Rhodes
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#*********************************************************************/

from PyQt6.QtWidgets import QGridLayout, QWidget, QSlider, QLabel
from PyQt6.QtCore import Qt
from cayenue.enums import ProxyType

class ImageTab(QWidget):
    def __init__(self, cp):
        super().__init__()
        self.cp = cp

        self.sldBrightness = QSlider(Qt.Orientation.Horizontal)
        self.sldBrightness.valueChanged.connect(cp.onEdit)
        self.sldSaturation = QSlider(Qt.Orientation.Horizontal)
        self.sldSaturation.valueChanged.connect(cp.onEdit)
        self.sldContrast = QSlider(Qt.Orientation.Horizontal)
        self.sldContrast.valueChanged.connect(cp.onEdit)
        self.sldSharpness = QSlider(Qt.Orientation.Horizontal)
        self.sldSharpness.valueChanged.connect(cp.onEdit)

        lblBrightness = QLabel("Brightness")
        lblSaturation = QLabel("Saturation")
        lblContrast = QLabel("Contrast")
        lblSharpness = QLabel("Sharpness")

        lytMain = QGridLayout(self)
        lytMain.addWidget(lblBrightness,      0, 0, 1, 1)
        lytMain.addWidget(self.sldBrightness, 0, 1, 1, 3)        
        lytMain.addWidget(lblSaturation,      1, 0, 1, 1)
        lytMain.addWidget(self.sldSaturation, 1, 1, 1, 3)        
        lytMain.addWidget(lblContrast,        2, 0, 1, 1)
        lytMain.addWidget(self.sldContrast,   2, 1, 1, 3)        
        lytMain.addWidget(lblSharpness,       3, 0, 1, 1)
        lytMain.addWidget(self.sldSharpness,  3, 1, 1, 3)        
        
    def fill(self, onvif_data):
        self.sldBrightness.setMaximum(onvif_data.brightness_max())
        self.sldBrightness.setMinimum(onvif_data.brightness_min())
        self.sldBrightness.setValue(onvif_data.brightness())

        self.sldContrast.setMaximum(onvif_data.contrast_max())
        self.sldContrast.setMinimum(onvif_data.contrast_min())
        self.sldContrast.setValue(onvif_data.contrast())

        self.sldSaturation.setMaximum(onvif_data.saturation_max())
        self.sldSaturation.setMinimum(onvif_data.saturation_min())
        self.sldSaturation.setValue(onvif_data.saturation())

        self.sldSharpness.setMaximum(onvif_data.sharpness_max())
        self.sldSharpness.setMinimum(onvif_data.sharpness_min())
        self.sldSharpness.setValue(onvif_data.sharpness())

        self.setEnabled(True)
        self.cp.onEdit()

    def edited(self, onvif_data):
        result = False
        if self.isEnabled():
            if not onvif_data.brightness() == self.sldBrightness.value():
                result = True
            if not onvif_data.contrast() == self.sldContrast.value():
                result = True
            if not onvif_data.saturation() == self.sldSaturation.value():
                result = True
            if not onvif_data.sharpness() == self.sldSharpness.value():
                result = True

        return result
    
    def update(self, onvif_data):
        if self.edited(onvif_data):
            onvif_data.setBrightness(self.sldBrightness.value())
            onvif_data.setSaturation(self.sldSaturation.value())
            onvif_data.setContrast(self.sldContrast.value())
            onvif_data.setSharpness(self.sldSharpness.value())
            if self.cp.mw.settingsPanel.proxy.proxyType == ProxyType.CLIENT:
                arg = "UPDATE IMAGE\n\n" + onvif_data.toJSON() + "\r\n"
                self.cp.mw.client.transmit(bytearray(arg, 'utf-8'))
            else:
                onvif_data.startUpdateImage()




#/********************************************************************
# Cayenue/panels/audio/modules/sample.py 
#
# Copyright (c) 2023  Stephen Rhodes
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#*********************************************************************/

import numpy as np
from numpy.fft import fft
import math
from loguru import logger
from PyQt6.QtWidgets import QGridLayout, QWidget, QSlider, QLabel, QWidget, QCheckBox
from PyQt6.QtGui import QPainter, QColorConstants, QColor
from PyQt6.QtCore import QPointF, Qt, QRectF
from cayenue.components import WarningBar, Indicator

MODULE_NAME = "sample"

class SampleSettings():
    def __init__(self, mw, camera=None):
        self.camera = camera
        self.mw = mw
        self.id = "File"
        if camera:
            self.id = camera.serial_number()

        self.amplitudeEnabled = self.getAudioAmplitudeEnabled()
        self.amplitudeGain = self.getAudioAmplitudeGain()
        self.frequencyEnabled = self.getAudioFrequencyEnabled()
        self.frequencyGain = self.getAudioFrequencyGain()
        self.frequencyPctHighPass = self.getAudioFrequencyPctHighPass()
        self.frequencyPctLowPass = self.getAudioFrequencyPctLowPass()
        self.frequencyPctCoverage = self.getAudioFrequencyPctCoverage()
        
    def getAudioAmplitudeEnabled(self):
        key = f'{self.id}/{MODULE_NAME}AudioAmplitudeEnabled'
        return bool(int(self.mw.settings.value(key, 1)))
    
    def setAudioAmplitudeEnabled(self, state):
        key = f'{self.id}/{MODULE_NAME}AudioAmplitudeEnabled'
        self.amplitudeEnabled = bool(state)
        self.mw.settings.setValue(key, int(state))

    def getAudioAmplitudeGain(self):
        key = f'{self.id}/{MODULE_NAME}AudioAmplitudeGain'
        return int(self.mw.settings.value(key, 50))
    
    def setAudioAmplitudeGain(self, value):
        key = f'{self.id}/{MODULE_NAME}AudioAmplitudeGain'
        self.amplitudeGain = value
        self.mw.settings.setValue(key, value)

    def getAudioFrequencyEnabled(self):
        key = f'{self.id}/{MODULE_NAME}AudioFrequencyEnabled'
        return bool(int(self.mw.settings.value(key, 1)))
    
    def setAudioFrequencyEnabled(self, state):
        key = f'{self.id}/{MODULE_NAME}AudioFrequencyEnabled'
        self.frequencyEnabled = bool(state)
        self.mw.settings.setValue(key, int(state))

    def getAudioFrequencyGain(self):
        key = f'{self.id}/{MODULE_NAME}AudioFrequencyGain'
        return int(self.mw.settings.value(key, 50))
    
    def setAudioFrequencyGain(self, value):
        key = f'{self.id}/{MODULE_NAME}AudioFrequencyGain'
        self.frequencyGain = value
        self.mw.settings.setValue(key, value)
    
    def getAudioFrequencyPctHighPass(self):
        key = f'{self.id}/{MODULE_NAME}AudioFrequencyPctHighPass'
        return float(self.mw.settings.value(key, 0))
    
    def setAudioFrequencyPctHighPass(self, value):
        key = f'{self.id}/{MODULE_NAME}AudioFrequencyPctHighPass'
        self.frequencyPctHighPass = value
        self.mw.settings.setValue(key, value)

    def getAudioFrequencyPctLowPass(self):
        key = f'{self.id}/{MODULE_NAME}AudioFrequencyPctLowPass'
        return float(self.mw.settings.value(key, 1))
    
    def setAudioFrequencyPctLowPass(self, value):
        key = f'{self.id}/{MODULE_NAME}AudioFrequencyPctLowPass'
        self.frequencyPctLowPass = value
        self.mw.settings.setValue(key, value)

    def getAudioFrequencyPctCoverage(self):
        key = f'{self.id}/{MODULE_NAME}AudioFrequencyPctCoverage'
        return float(self.mw.settings.value(key, 1))
    
    def setAudioFrequencyPctCoverage(self, value):
        key = f'{self.id}/{MODULE_NAME}AudioFrequencyPctCoverage'
        self.frequencyPctCoverage = value
        self.mw.settings.setValue(key, value)

class AmplitudeDisplay(QLabel):
    def __init__(self, mw):
        super().__init__()
        self.mw = mw
        self.data = None
        self.lock = False
        self.setStyleSheet("QLabel { background : black; }")
    
    def paintEvent(self, event):
        last_point = None
        width = self.size().width()
        height = self.size().height()
        painter = QPainter(self)
        painter.setPen(QColorConstants.DarkGreen)

        if self.data is not None:
            x_unit = width / self.data.size
            for i, value in enumerate(self.data):
                x = i * x_unit
                y = height*(1 - (1 + value)/2)
                current_point = QPointF(x, y)
                if last_point and current_point:
                    painter.drawLine(last_point, current_point)
                last_point = current_point

        if not self.mw.audioConfigure.chkAmplitude.isChecked() or not self.isEnabled() or self.data is None:
            painter.eraseRect(0, 0, width, height)

    def setData(self, data):
        self.data = data
        self.update()

class FrequencyDisplay(QLabel):
    def __init__(self, mw):
        super().__init__()
        self.mw = mw
        self.data = None
        self.lock = False
        self.setStyleSheet("QLabel { background : black; }")
        self.setMouseTracking(True)
        self.draggingLowPass = False
        self.draggingHighPass = False
        self.pctCoverage = 1.0
        self.pctHighPass = 0
        self.pctLowPass = 1.0
        self.handleLowPass = QRectF(0, 0, 8, 8)
        self.handleHighPass = QRectF(0, 0, 8, 8)

    def paintEvent(self, event):
        width = self.size().width()
        height = self.size().height()
        painter = QPainter(self)
        painter.setPen(QColorConstants.DarkGreen)
        lowPass = int(self.pctLowPass * width)
        highPass = int(self.pctHighPass * width)

        if self.data is not None:
            x_unit = width / self.data.size
            # the y_scale is arbitrary based on esthetics
            y_scale = height/4 
            for i, value in enumerate(self.data):
                x = i * x_unit
                y = height - (value * y_scale)
                painter.setPen(QColorConstants.DarkGreen)
                if highPass < lowPass:
                    if x < highPass or x > lowPass:
                        painter.setPen(QColor("#303030"))
                else:
                    if x < highPass and x > lowPass:
                        painter.setPen(QColor("#303030"))
                painter.drawLine(QPointF(x, height), QPointF(x, y))

        if not self.mw.audioConfigure.chkFrequency.isChecked() or not self.isEnabled() or self.data is None:
            painter.eraseRect(0, 0, width, height)

        self.handleLowPass.moveCenter(QPointF(lowPass, height/2))
        self.handleHighPass.moveCenter(QPointF(highPass, height/2))
        painter.setPen(QColorConstants.DarkYellow)
        painter.drawRect(self.handleLowPass)
        painter.drawRect(self.handleHighPass)
        painter.drawLine(QPointF(lowPass, 0), QPointF(lowPass, self.handleLowPass.top()))
        painter.drawLine(QPointF(lowPass, self.handleLowPass.bottom()), QPointF(lowPass, height-1))
        painter.drawLine(QPointF(highPass, 0), QPointF(highPass, self.handleHighPass.top()))
        painter.drawLine(QPointF(highPass, self.handleHighPass.bottom()), QPointF(highPass, height-1))

    def mouseMoveEvent(self, event):
        if self.handleLowPass.contains(QPointF(event.pos())) or self.handleHighPass.contains(QPointF(event.pos())):
            self.setCursor(Qt.CursorShape.OpenHandCursor)
        else:
            self.setCursor(Qt.CursorShape.ArrowCursor)

        lowPass = int(self.pctLowPass * self.width())
        if self.draggingLowPass:
            x = min(max(event.pos().x(), 0), self.size().width()-1)
            lowPass = x
            self.pctLowPass = lowPass / self.width()
            if camera := self.mw.cameraPanel.getCurrentCamera():
                camera.audioModelSettings.setAudioFrequencyPctLowPass(self.pctLowPass)
            self.update()

        highPass = int(self.pctHighPass * self.width())
        if self.draggingHighPass:
            x = min(max(event.pos().x(), 0), self.size().width())
            highPass = x
            self.pctHighPass = highPass / self.width()
            if camera := self.mw.cameraPanel.getCurrentCamera():
                camera.audioModelSettings.setAudioFrequencyPctHighPass(self.pctHighPass)
            self.update()

        if self.draggingHighPass or self.draggingLowPass:
            if highPass < lowPass:
                self.pctCoverage = 1 - (self.width() - lowPass + highPass - 1) / self.width()
            else:
                self.pctCoverage = (highPass - lowPass - 1) / self.width()
            self.pctCoverage = max(self.pctCoverage, 0.01)
            if camera := self.mw.cameraPanel.getCurrentCamera():
                camera.audioModelSettings.setAudioFrequencyPctCoverage(self.pctCoverage)

        return super().mouseMoveEvent(event)
    
    def mousePressEvent(self, event):
        if self.handleLowPass.contains(QPointF(event.pos())):
            self.setCursor(Qt.CursorShape.ClosedHandCursor)
            self.draggingLowPass = True

        if self.handleHighPass.contains(QPointF(event.pos())):
            self.setCursor(Qt.CursorShape.ClosedHandCursor)
            self.draggingHighPass = True

        return super().mousePressEvent(event)
    
    def mouseReleaseEvent(self, event):
        self.setCursor(Qt.CursorShape.OpenHandCursor)
        self.draggingLowPass = False
        self.draggingHighPass = False
        return super().mouseReleaseEvent(event)
    
    def setData(self, data):
        self.data = data
        self.update()

    def setPctHighPass(self, value):
        self.pctHighPass = value
        self.highPass = max(int(value * self.width()), 0)

    def setPctLowPass(self, value):
        self.pctLowPass = value
        self.lowPass = min(int(value * self.width()), self.width() - 1)

class AudioConfigure(QWidget):
    def __init__(self, mw):
        try:
            super().__init__()
            self.mw = mw
            self.name = MODULE_NAME
            self.amplitudeGain = 1.0
            self.frequencyGain = 1.0

            self.dspAmplitude = AmplitudeDisplay(self.mw)
            self.chkAmplitude = QCheckBox("Amplitude")
            self.chkAmplitude.setChecked(True)
            self.chkAmplitude.stateChanged.connect(self.chkAmpStateChanged)
            
            self.sldAmplitudeGain = QSlider(Qt.Orientation.Vertical)
            self.sldAmplitudeGain.setValue(50)
            self.sldAmplitudeGain.valueChanged.connect(self.sldAmpGainValueChanged)

            self.barAmplitude = WarningBar()
            self.indAmplitude = Indicator(self.mw)
            self.lblAmplitudeGain = QLabel("Gain")
            self.dspFrequency = FrequencyDisplay(self.mw)
            self.chkFrequency = QCheckBox("Frequency")
            self.chkFrequency.setChecked(True)
            self.chkFrequency.stateChanged.connect(self.chkFreqStateChanged)

            self.sldFrequencyGain = QSlider(Qt.Orientation.Vertical)
            self.sldFrequencyGain.setValue(50)
            self.sldFrequencyGain.valueChanged.connect(self.sldFreqGainValueChanged)
            self.lblFrequencyGain = QLabel("Gain")
            self.barFrequency = WarningBar()
            self.indFrequency = Indicator(self.mw)

            self.pnlAmplitude = QWidget()
            lytAmplitude = QGridLayout(self.pnlAmplitude)
            lytAmplitude.addWidget(self.chkAmplitude,       0, 0, 1, 1, Qt.AlignmentFlag.AlignCenter)
            lytAmplitude.addWidget(self.dspAmplitude,       1, 0, 4, 1)
            lytAmplitude.addWidget(self.sldAmplitudeGain,   2, 3, 1, 1, Qt.AlignmentFlag.AlignHCenter)
            lytAmplitude.addWidget(self.lblAmplitudeGain,   3, 3, 1, 1)
            lytAmplitude.addWidget(self.indAmplitude,       1, 4, 1, 1)
            lytAmplitude.addWidget(self.barAmplitude,       2, 4, 2, 1)
            lytAmplitude.setColumnStretch(0, 10)
            lytAmplitude.setColumnStretch(4, 2)
            lytAmplitude.setContentsMargins(0, 0, 0, 0)

            self.pnlFrequency = QWidget()
            lytFrequency = QGridLayout(self.pnlFrequency)
            lytFrequency.addWidget(self.chkFrequency,       0, 0, 1, 1, Qt.AlignmentFlag.AlignCenter)
            lytFrequency.addWidget(self.dspFrequency,       1, 0, 4, 1)
            lytFrequency.addWidget(self.sldFrequencyGain,   2, 3, 1, 1, Qt.AlignmentFlag.AlignHCenter)
            lytFrequency.addWidget(self.lblFrequencyGain,   3, 3, 1, 1)
            lytFrequency.addWidget(self.indFrequency,       1, 4, 1, 1)
            lytFrequency.addWidget(self.barFrequency,       2, 4, 2, 1)
            lytFrequency.setColumnStretch(0, 10)
            lytFrequency.setColumnStretch(4, 2)
            lytFrequency.setContentsMargins(0, 0, 0, 0)

            self.pnlAmplitude.setMinimumHeight(200)
            self.pnlFrequency.setMinimumHeight(200)

            lytMain = QGridLayout(self)
            lytMain.addWidget(self.pnlAmplitude,      0, 0, 1, 4)
            lytMain.addWidget(self.pnlFrequency,      1, 0, 1, 4)
            lytMain.addWidget(QLabel(),               3, 0, 1, 4)
            lytMain.setRowStretch(3, 3)

            self.enableControls(False)

        except:
            logger.exception("sample configuration failed to load")

    def sldAmpGainValueChanged(self, value):
        self.amplitudeGain = math.exp(0.1 * (value - 50))
        if camera := self.mw.cameraPanel.getCurrentCamera():
            camera.audioModelSettings.setAudioAmplitudeGain(value)

    def sldFreqGainValueChanged(self, value):
        self.frequencyGain = math.exp(0.05 * (value - 50))
        if camera := self.mw.cameraPanel.getCurrentCamera():
            camera.audioModelSettings.setAudioFrequencyGain(value)

    def chkAmpStateChanged(self, state):
        self.barAmplitude.setLevel(0.0)
        self.indAmplitude.setState(0)
        self.dspAmplitude.setData(None)
        if camera := self.mw.cameraPanel.getCurrentCamera():
            camera.audioModelSettings.setAudioAmplitudeEnabled(state)

    def chkFreqStateChanged(self, state):
        self.barFrequency.setLevel(0.0)
        self.indFrequency.setState(0)
        self.dspFrequency.setData(None)
        if camera := self.mw.cameraPanel.getCurrentCamera():
            camera.audioModelSettings.setAudioFrequencyEnabled(state)

    def clearIndicators(self):
        self.barAmplitude.setLevel(0)
        self.indAmplitude.setState(0)
        self.barFrequency.setLevel(0)
        self.indFrequency.setState(0)

    def enableControls(self, state):
        self.clearIndicators()
        self.pnlAmplitude.setEnabled(bool(state))
        self.pnlFrequency.setEnabled(bool(state))

    def setCamera(self, camera):
        if not camera: return

        if not self.isModelSettings(camera.audioModelSettings):
            camera.audioModelSettings = SampleSettings(self.mw, camera)

        self.mw.audioPanel.lblCamera.setText(f'Camera - {camera.name()}')
        self.chkAmplitude.setChecked(camera.audioModelSettings.amplitudeEnabled)
        self.sldAmplitudeGain.setValue(camera.audioModelSettings.amplitudeGain)
        self.chkFrequency.setChecked(camera.audioModelSettings.frequencyEnabled)
        self.sldFrequencyGain.setValue(camera.audioModelSettings.frequencyGain)
        self.dspFrequency.setPctHighPass(camera.audioModelSettings.frequencyPctHighPass)
        self.dspFrequency.setPctLowPass(camera.audioModelSettings.frequencyPctLowPass)
        self.barAmplitude.setLevel(0)
        self.indAmplitude.setState(0)
        self.barFrequency.setLevel(0)
        self.indFrequency.setState(0)

        if not camera.hasAudio():
            self.chkAmplitude.setChecked(False)
            self.chkFrequency.setChecked(False)

        enable = True
        profile = self.mw.cameraPanel.getProfile(camera.uri())
        if profile:
            if profile.getDisableAudio() or not profile.getAnalyzeAudio():
                enable = False
        if not camera.hasAudio():
            enable = False
        self.enableControls(enable)

        self.update()

    def isModelSettings(self, arg):
        return type(arg) == SampleSettings

class AudioWorker:
    def __init__(self, mw):
        try:
            self.mw = mw
            self.last_error = ""
        except:
            logger.exception("sample worker failed to load")

    def __call__(self, F, player):
        try:
            if not self.mw.audioConfigure:
                return

            if not F or not player:
                self.mw.audioConfigure.dspAmplitude.setData(None)
                self.mw.audioConfigure.dspFrequency.setData(None)
                self.mw.audioConfigure.clearIndicators()
                return
            
            if player.getAudioCodec() != "aac":
                raise Exception("Unsupported audio codec, only AAC is supported")


            samples = np.array(F, copy=True)
            if F.channels() > 1:
                samples = samples[::F.channels()]

            if self.mw.audioConfigure.name != MODULE_NAME:
                return
            
            player.lock()
            camera = self.mw.cameraPanel.getCamera(player.uri)
            if not self.mw.audioConfigure.isModelSettings(player.audioModelSettings):
                if player.isCameraStream():
                    if camera:
                        camera.audioModelSettings = SampleSettings(self.mw, camera)
                    player.audioModelSettings = camera.audioModelSettings

            if not player.audioModelSettings:
                raise Exception("Unable to set audio model parameters for player")

            spectrum = None
            if player.audioModelSettings.frequencyEnabled and samples.size:
                spectrum = fft(samples)

            if player.audioModelSettings.amplitudeEnabled:
                samples *= math.exp(0.2 * (player.audioModelSettings.amplitudeGain - 50))
                rms = 0
                if samples.size:
                    for s in samples:
                        rms += s * s
                    rms = math.sqrt(rms/samples.size)

                alarmState = False
                if rms > 1:
                    alarmState = True

                if camera := self.mw.cameraPanel.getCurrentCamera():
                    if camera.isCurrent():
                        self.mw.audioConfigure.barAmplitude.setLevel(rms)
                        self.mw.audioConfigure.dspAmplitude.setData(samples)
                        if alarmState:
                            self.mw.audioConfigure.indAmplitude.setState(1)

                player.handleAlarm(alarmState)

            if player.audioModelSettings.frequencyEnabled and spectrum is not None:
                half_length = int(len(spectrum)/2)
                frequencies = np.abs(spectrum[:half_length])
                frequencies *= math.exp(0.05 * (player.audioModelSettings.frequencyGain - 50))

                highPass = player.audioModelSettings.frequencyPctHighPass * frequencies.size
                lowPass = player.audioModelSettings.frequencyPctLowPass * frequencies.size
                sph = 0
                for i, f in enumerate(frequencies):
                    if highPass < lowPass:
                        if i > highPass and i < lowPass:
                            sph += f
                    else:
                        if i > highPass or i < lowPass:
                            sph += f
                sph /= (frequencies.size * player.audioModelSettings.frequencyPctCoverage)

                alarmState = False
                if sph > 1:
                    alarmState = True

                if camera := self.mw.cameraPanel.getCurrentCamera():
                    if camera.isCurrent():
                        self.mw.audioConfigure.barFrequency.setLevel(sph)
                        self.mw.audioConfigure.dspFrequency.setData(frequencies)
                        if alarmState:
                            self.mw.audioConfigure.indFrequency.setState(1)

                player.handleAlarm(alarmState)
                self.mw.audioPanel.lblMessage.setText("")
                self.last_error = ""

        except Exception as err:
            if str(err) != self.last_error:
                logger.exception("pyAudioCallback exception")
                self.mw.audioPanel.lblMessage.setText(str(err))
            self.last_error = str(err)
        player.unlock()




#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# Copyright (c) Megvii Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#******************************************************************************/

import torch
import torch.nn as nn


class IOUloss(nn.Module):
    def __init__(self, reduction="none", loss_type="iou"):
        super(IOUloss, self).__init__()
        self.reduction = reduction
        self.loss_type = loss_type

    def forward(self, pred, target):
        assert pred.shape[0] == target.shape[0]

        pred = pred.view(-1, 4)
        target = target.view(-1, 4)
        tl = torch.max(
            (pred[:, :2] - pred[:, 2:] / 2), (target[:, :2] - target[:, 2:] / 2)
        )
        br = torch.min(
            (pred[:, :2] + pred[:, 2:] / 2), (target[:, :2] + target[:, 2:] / 2)
        )

        area_p = torch.prod(pred[:, 2:], 1)
        area_g = torch.prod(target[:, 2:], 1)

        en = (tl < br).type(tl.type()).prod(dim=1)
        area_i = torch.prod(br - tl, 1) * en
        area_u = area_p + area_g - area_i
        iou = (area_i) / (area_u + 1e-16)

        if self.loss_type == "iou":
            loss = 1 - iou ** 2
        elif self.loss_type == "giou":
            c_tl = torch.min(
                (pred[:, :2] - pred[:, 2:] / 2), (target[:, :2] - target[:, 2:] / 2)
            )
            c_br = torch.max(
                (pred[:, :2] + pred[:, 2:] / 2), (target[:, :2] + target[:, 2:] / 2)
            )
            area_c = torch.prod(c_br - c_tl, 1)
            giou = iou - (area_c - area_u) / area_c.clamp(1e-16)
            loss = 1 - giou.clamp(min=-1.0, max=1.0)

        if self.reduction == "mean":
            loss = loss.mean()
        elif self.reduction == "sum":
            loss = loss.sum()

        return loss



#/********************************************************************
# Cayenue/panels/video/modules/motion.py 
#
# Copyright (c) 2023  Stephen Rhodes
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#*********************************************************************/

import math
import numpy as np
import cv2
from PyQt6.QtWidgets import QGridLayout, QWidget, QSlider, QGroupBox, QLabel
from PyQt6.QtCore import Qt
from loguru import logger
from cayenue.components import WarningBar, Indicator
from cayenue.panels.video.modules.settings import VideoModelSettings

MODULE_NAME = "motion"

class VideoConfigure(QWidget):
    def __init__(self, mw):
        try:
            super().__init__()
            self.mw = mw
            self.name = MODULE_NAME
            self.source = None
            self.camera = None
            self.initialized = False
            
            self.barLevel = WarningBar()
            self.indAlarm = Indicator(self.mw)
            self.sldGain = QSlider(Qt.Orientation.Vertical)
            self.sldGain.setValue(50)
            self.sldGain.valueChanged.connect(self.sldGainValueChanged)
            grpSlide = QGroupBox()
            grpSlide.setMaximumHeight(260)
        
            lytSlide = QGridLayout(grpSlide)
            lytSlide.addWidget(self.sldGain,   1, 0, 1, 1, Qt.AlignmentFlag.AlignHCenter)
            lytSlide.addWidget(QLabel("Gain"), 2, 0, 1, 1, Qt.AlignmentFlag.AlignCenter)
            lytSlide.addWidget(self.indAlarm,  0, 1, 1, 1)
            lytSlide.addWidget(self.barLevel,  1, 1, 2, 1)
            lytSlide.addWidget(QLabel(),       0, 2, 1, 1)
            lytSlide.setRowStretch(1, 10)
            
            lytMain = QGridLayout(self)
            lytMain.addWidget(grpSlide,      0, 0, 1, 1)
            lytMain.addWidget(QLabel(),      1, 0, 1, 1)

            self.enableControls(False)
            if camera := self.mw.cameraPanel.getCurrentCamera():
                self.setCamera(camera)
            else:
                if file := self.mw.filePanel.getCurrentFileURI():
                    self.setFile(file)
            self.initialized = True

        except:
            logger.exception("sample configuration failed to load")

    def sldGainValueChanged(self, value):
        if camera := self.mw.cameraPanel.getCurrentCamera():
            camera.videoModelSettings.setModelOutputGain(value)

    def setCamera(self, camera):
        if not camera:
            return
        
        self.camera = camera
        if camera.videoModelSettings.module_name != MODULE_NAME:
            camera.videoModelSettings = VideoModelSettings(self.mw, camera, "motion")
        self.mw.videoPanel.lblCamera.setText(f'Camera - {camera.name()}')
        self.sldGain.setValue(camera.videoModelSettings.gain)
        self.barLevel.setLevel(0)
        self.indAlarm.setState(0)
        profile = self.mw.cameraPanel.getProfile(camera.uri())
        if profile:
            self.enableControls(profile.getAnalyzeVideo())

    def enableControls(self, state):
        self.setEnabled(bool(state))

class VideoWorker:
    def __init__(self, mw):
        try:
            self.mw = mw
            self.last_ex = ""
            self.last_img = None
            self.first_pass = True
            self.kernel = np.array((9, 9), dtype=np.uint8)
            self.last_alarm_state = False

        except:
            logger.exception("sample worker failed to load")

    def __call__(self, F, uri):
        try:
            if not self.mw.videoConfigure:
                return

            player = self.mw.pm.getPlayer(uri)
            if not player:
                return
            
            camera = self.mw.cameraPanel.getCamera(uri)
            if not camera:
                return

            if not len(F) or not player or self.mw.videoConfigure.name != MODULE_NAME:
                self.mw.videoConfigure.barLevel.setLevel(0)
                self.mw.videoConfigure.indAlarm.setState(0)
                return
            
            if camera.videoModelSettings.module_name != MODULE_NAME:
                camera.videoModelSettings = VideoModelSettings(self.mw, camera, MODULE_NAME)

            img = np.array(F, copy = False)
            img = cv2.cvtColor(img, cv2.COLOR_RGB2GRAY)

            level = 0
            diff = img
            if player.last_image is not None:
                diff = cv2.subtract(img, player.last_image)
                diff = cv2.medianBlur(diff, 3)
                diff = cv2.adaptiveThreshold(diff, 255, cv2.ADAPTIVE_THRESH_GAUSSIAN_C, cv2.THRESH_BINARY_INV, 11, 3)
                diff = cv2.medianBlur(diff, 3)
                diff = cv2.morphologyEx(diff, cv2.MORPH_CLOSE, self.kernel, iterations=1)

                motion = diff.sum() / (diff.shape[0] * diff.shape[1])
                level = math.exp(0.2 * (camera.videoModelSettings.gain - 50)) * motion

            player.last_image = img

            alarmState = level > 1.0

            if player.uri == self.mw.glWidget.focused_uri:
                self.mw.videoConfigure.barLevel.setLevel(level)
                if alarmState:
                    self.mw.videoConfigure.indAlarm.setState(1)

            player.handleAlarm(alarmState)

        except Exception as ex:
            if self.last_ex != str(ex) and self.mw.videoConfigure.name == MODULE_NAME:
                logger.exception(f'Video worker motion detector call error {ex}')
            self.last_ex = str(ex)


from .fileselector import FileSelector
from .directoryselector import DirectorySelector
from .thresholdslider import ThresholdSlider
from .comboselector import ComboSelector
from .progress import Progress
from .warningbar import WarningBar, Indicator
from .target import Target, TargetList, TargetDialog, TargetSelector
from .diskmanager import DiskManager
from .infodialog import InfoDialog


#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# Copyright (c) Megvii Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#******************************************************************************/

import torch
import torch.nn as nn

from .darknet import Darknet
from .network_blocks import BaseConv


class YOLOFPN(nn.Module):
    """
    YOLOFPN module. Darknet 53 is the default backbone of this model.
    """

    def __init__(
        self,
        depth=53,
        in_features=["dark3", "dark4", "dark5"],
    ):
        super().__init__()

        self.backbone = Darknet(depth)
        self.in_features = in_features

        # out 1
        self.out1_cbl = self._make_cbl(512, 256, 1)
        self.out1 = self._make_embedding([256, 512], 512 + 256)

        # out 2
        self.out2_cbl = self._make_cbl(256, 128, 1)
        self.out2 = self._make_embedding([128, 256], 256 + 128)

        # upsample
        self.upsample = nn.Upsample(scale_factor=2, mode="nearest")

    def _make_cbl(self, _in, _out, ks):
        return BaseConv(_in, _out, ks, stride=1, act="lrelu")

    def _make_embedding(self, filters_list, in_filters):
        m = nn.Sequential(
            *[
                self._make_cbl(in_filters, filters_list[0], 1),
                self._make_cbl(filters_list[0], filters_list[1], 3),
                self._make_cbl(filters_list[1], filters_list[0], 1),
                self._make_cbl(filters_list[0], filters_list[1], 3),
                self._make_cbl(filters_list[1], filters_list[0], 1),
            ]
        )
        return m

    def load_pretrained_model(self, filename="./weights/darknet53.mix.pth"):
        with open(filename, "rb") as f:
            state_dict = torch.load(f, map_location="cpu")
        print("loading pretrained weights...")
        self.backbone.load_state_dict(state_dict)

    def forward(self, inputs):
        """
        Args:
            inputs (Tensor): input image.

        Returns:
            Tuple[Tensor]: FPN output features..
        """
        #  backbone
        out_features = self.backbone(inputs)
        x2, x1, x0 = [out_features[f] for f in self.in_features]

        #  yolo branch 1
        x1_in = self.out1_cbl(x0)
        x1_in = self.upsample(x1_in)
        x1_in = torch.cat([x1_in, x1], 1)
        out_dark4 = self.out1(x1_in)

        #  yolo branch 2
        x2_in = self.out2_cbl(out_dark4)
        x2_in = self.upsample(x2_in)
        x2_in = torch.cat([x2_in, x2], 1)
        out_dark3 = self.out2(x2_in)

        outputs = (out_dark3, out_dark4, x0)
        return outputs



#*******************************************************************************
# Cayenue/cayenue/components/fileselector.py
#
# Copyright (c) 2023 Stephen Rhodes 
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#******************************************************************************/

from PyQt6.QtWidgets import QWidget, QLineEdit, QPushButton, \
    QLabel, QGridLayout, QFileDialog
import platform

class FileSelector(QWidget):
    def __init__(self, mw, name):
        super().__init__()
        self.mw = mw
        self.filenameKey = "Module/" + name + "/filename"

        self.txtFilename = QLineEdit()
        self.txtFilename.setText(self.mw.settings.value(self.filenameKey))
        self.txtFilename.textEdited.connect(self.txtFilenameChanged)
        self.btnSelect = QPushButton("...")
        self.btnSelect.clicked.connect(self.btnSelectClicked)
        self.lblSelect = QLabel("Model")

        lytMain = QGridLayout(self)
        lytMain.addWidget(self.lblSelect,     0, 0, 1, 1)
        lytMain.addWidget(self.txtFilename,   0, 1, 1, 1)
        lytMain.addWidget(self.btnSelect,     0, 2, 1, 1)
        lytMain.setColumnStretch(1, 10)
        lytMain.setContentsMargins(0, 0, 0, 0)

    def btnSelectClicked(self):
        filename = None
        if platform.system() == "Linux":
            filename = QFileDialog.getOpenFileName(self, "Select File", self.txtFilename.text(), options=QFileDialog.Option.DontUseNativeDialog)[0]
        else:
            filename = QFileDialog.getOpenFileName(self, "Select File", self.txtFilename.text())[0]

        if filename:
            self.txtFilename.setText(filename)
            self.mw.settings.setValue(self.filenameKey, filename)

    def txtFilenameChanged(self, text):
        self.mw.settings.setValue(self.filenameKey, text)

    def text(self):
        return self.txtFilename.text()



#********************************************************************
# Cayenue/cayenue/enums.py
#
# Copyright (c) 2025  Stephen Rhodes
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#*********************************************************************/

from enum import Enum

class ProxyType(Enum):
    STAND_ALONE = 0
    SERVER = 1
    CLIENT = 2

class StreamState(Enum):
    IDLE = 0
    CONNECTING = 1
    CONNECTED = 2
    INVALID = 3

class Style(Enum):
    DARK = 0
    LIGHT = 1

class Occurence(Enum):
    BEFORE = 0
    DURING = 1
    AFTER = 2

class SnapshotAuth(Enum):
    UNKNOWN = 0
    NOT_WORKING = 1
    BASIC = 2
    DIGEST = 3


#*******************************************************************************
# Cayenue/cayenue/components/diskmanager.py
#
# Copyright (c) 2026 Stephen Rhodes 
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#******************************************************************************/
import os
from datetime import datetime
from pathlib import Path
from loguru import logger
from PyQt6.QtCore import QFileInfo

class FileInfo:
    def __init__(self, path: Path):
        self.path = path
        self.name = path.name
        self.size = path.stat().st_size
        self.modified_time = datetime.fromtimestamp(path.stat().st_mtime)
        self.created_time = datetime.fromtimestamp(path.stat().st_ctime)

def __repr__(self):
        return (f"FileInfo(path='{self.path}', name='{self.name}', created='{self.created_time}', "
                f"modified='{self.modified_time}', size={self.size})")

class DiskManager():
    def __init__(self, mw):
        self.mw = mw

    def list_files(self, directory: str) -> tuple[list[FileInfo], int]:
        total_size = 0
        file_infos = []
        for dirpath, dirnames, filenames in os.walk(directory):
            for f in filenames:
                try:
                    file_info = FileInfo(Path(os.path.join(dirpath, f)))
                    total_size += file_info.size
                    file_infos.append(file_info)
                except Exception as ex:
                    print(f"file listing exception: {ex}")
                    pass

        file_infos.sort(key=lambda x: x.created_time)
        return file_infos, total_size

    def removeAssociatedPictureFiles(self, filename):
        try:
            info = QFileInfo(filename)
            finish = info.lastModified()
            dir = info.absoluteDir().dirName()
            pic_dir = os.path.join(self.mw.settingsPanel.storage.dirPictures.txtDirectory.text(), dir)
            files = os.listdir(pic_dir)
            for file in files:
                pic_info = QFileInfo(os.path.join(pic_dir, file))
                stem = Path(file).stem
                if len(stem) == 14 and stem.isnumeric():
                    target = pic_info.birthTime()
                    #if (target >= start and target <= finish):
                    # just wipe all the older pictures
                    if target <= finish:
                        os.remove(os.path.join(pic_dir, file))
        except Exception as ex:
            logger.error(f'Exception occurred during removal of associated picture files: {ex}')

    def manageDirectory(self, d):
        print("manage directory")

        files, size = self.list_files(d)
        max_size = self.mw.settingsPanel.storage.spnDiskLimit.value() * 1_000_000_000

        files_to_be_deleted = []
        total = 0
        diff = size - max_size
        if diff > 0:
            for file in files:
                total += file.size
                files_to_be_deleted.append(file)
                if total > diff:
                    break

        for file in files_to_be_deleted:
            try:
                self.removeAssociatedPictureFiles(str(file.path))
                os.remove(file.path)
            except Exception as ex:
                logger.error(f'File delete error: {ex}')
                pass
        
        self.mw.settingsPanel.storage.signals.updateDiskUsage.emit()
        print("DONE MANAGING DIRECTORY")



#/********************************************************************
# Cayenue/cayenue/panels/file/filepanel.py 
#
# Copyright (c) 2023  Stephen Rhodes
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#*********************************************************************/

import os
import sys
from PyQt6.QtWidgets import QGridLayout, QWidget, QLabel, \
    QMessageBox, QMenu, QApplication
from PyQt6.QtGui import QAction, QFileSystemModel
from PyQt6.QtCore import Qt, QStandardPaths
from cayenue.components import Progress, InfoDialog
from loguru import logger
import avio
from . import FileControlPanel
from cayenue.components.directoryselector import DirectorySelector
from . import TreeView

class FilePanel(QWidget):
    def __init__(self, mw):
        super().__init__()
        self.mw = mw
        self.alarmSoundVolume = 80
        self.expandedPaths = []
        self.loadedCount = 0
        self.restorationPath = None
        self.verticalScrollBarPosition = 0
        self.dlgInfo = InfoDialog(mw)

        if self.mw.parent_window:
            video_dir = self.mw.parent_window.settingsPanel.storage.dirArchive.text()
        else:
            video_dir = QStandardPaths.standardLocations(QStandardPaths.StandardLocation.MoviesLocation)[0]
        self.dirArchive = DirectorySelector(mw, self.mw.settingsPanel.storage.archiveKey, "", video_dir)
        #self.dirArchive.signals.dirChanged.connect(self.mw.settingsPanel.storage.dirArchiveChanged)
        self.dirArchive.signals.dirChanged.connect(self.dirChanged)

        self.model = QFileSystemModel()
        self.model.fileRenamed.connect(self.onFileRenamed)
        self.model.directoryLoaded.connect(self.loaded)
        self.tree = TreeView(mw)
        self.tree.setModel(self.model)
        self.tree.doubleClicked.connect(self.treeDoubleClicked)
        self.tree.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
        self.tree.customContextMenuRequested.connect(self.showContextMenu)

        self.progress = Progress(mw)
        self.control = FileControlPanel(mw, self)

        lytMain = QGridLayout(self)
        lytMain.addWidget(self.dirArchive,  0, 0, 1, 1)
        lytMain.addWidget(self.tree,       1, 0, 1, 1)
        lytMain.addWidget(self.progress,   2, 0, 1, 1)
        lytMain.addWidget(QLabel(),        3, 0, 1, 1)
        lytMain.addWidget(self.control,    4, 0, 1, 1)
        lytMain.setRowStretch(1, 10)

        self.dirChanged(self.dirArchive.text())

        self.menu = QMenu("Context Menu", self)
        self.remove = QAction("Delete", self)
        self.rename = QAction("Rename", self)
        self.info = QAction("Info", self)
        self.play = QAction("Play", self)
        self.stop = QAction("Stop", self)
        self.remove.triggered.connect(self.onMenuRemove)
        self.rename.triggered.connect(self.onMenuRename)
        self.info.triggered.connect(self.onMenuInfo)
        self.play.triggered.connect(self.onMenuPlay)
        self.stop.triggered.connect(self.onMenuStop)
        self.menu.addAction(self.remove)
        self.menu.addAction(self.rename)
        self.menu.addAction(self.info)
        self.menu.addAction(self.play)
        self.menu.addAction(self.stop)

    def loaded(self, path):
        self.loadedCount += 1
        self.model.sort(0)
        for i in range(self.model.rowCount(self.model.index(path))):
            idx = self.model.index(i, 0, self.model.index(path))
            if idx.isValid():
                if self.model.filePath(idx) in self.expandedPaths:
                    self.tree.setExpanded(idx, True)

        if len(self.expandedPaths):
            if self.loadedCount == len(self.expandedPaths) + 1:
                if self.verticalScrollBarPosition:
                    QApplication.processEvents()
                    self.tree.verticalScrollBar().setValue(self.verticalScrollBarPosition)
                self.expandedPaths.clear()
                self.restoreSelectedPath()
        else:
            self.restoreSelectedPath()

    def restoreSelectedPath(self):
        if self.restorationPath:
            idx = self.model.index(self.restorationPath)
            if idx.isValid():
                self.tree.setCurrentIndex(idx)
            self.restorationPath = None

    def refresh(self):
        self.loadedCount = 0
        self.expandedPaths = []
        self.restorationPath = self.model.filePath(self.tree.currentIndex())
        path = self.dirArchive.txtDirectory.text()
        self.model.sort(0)
        for i in range(self.model.rowCount(self.model.index(path))):
            idx = self.model.index(i, 0, self.model.index(path))
            if idx.isValid():
                if self.tree.isExpanded(idx):
                    self.expandedPaths.append(self.model.filePath(idx))
        self.verticalScrollBarPosition = self.tree.verticalScrollBar().value()
        self.model = QFileSystemModel()
        self.model.setRootPath(path)
        self.model.fileRenamed.connect(self.onFileRenamed)
        self.model.directoryLoaded.connect(self.loaded)
        self.tree.setModel(self.model)
        self.tree.setRootIndex(self.model.index(path))

    def dirChanged(self, path):
        if len(path) > 0:
            self.model.setRootPath(path)
            self.tree.setRootIndex(self.model.index(path))

    def treeDoubleClicked(self, index):
        if index.isValid():
            fileInfo = self.model.fileInfo(index)
            if fileInfo.isDir():
                self.tree.setExpanded(index, self.tree.isExpanded(index))
            else:
                self.mw.closeAnyPlayingFiles()
                if uri := self.getCurrentFileURI():
                    self.mw.playMedia(uri)
                    self.mw.glWidget.focused_uri = uri

    def onMediaStarted(self, duration):
        if self.mw.tab.currentIndex() == 1:
            self.tree.setFocus()
        self.control.setBtnPlay()
        self.control.setBtnMute()
        self.control.setSldVolume()

    def onMediaStopped(self, uri):
        self.control.setBtnPlay()
        self.progress.updateProgress(0.0)

        # this is needed to prevent overwriting another file duration
        another = None
        self.mw.pm.lock()
        for player in self.mw.pm.players.values():
            if not player.isCameraStream():
                another = player
                break
        self.mw.pm.unlock()
        if not another:
            self.progress.lblDuration.setText("0:00")

    def onMediaProgress(self, pct, uri):
        if player := self.mw.pm.getPlayer(uri):
            player.file_progress = pct
            self.progress.updateDuration(player.duration())

        #if pct >= 0.0 and pct <= 1.0 and uri == self.mw.glWidget.focused_uri:
        if pct >= 0.0 and pct <= 1.0:
            self.progress.updateProgress(pct)

    def showContextMenu(self, pos):
        player = self.mw.pm.getPlayer(self.getCurrentFileURI())
        self.remove.setDisabled(bool(player))
        self.rename.setDisabled(bool(player))
        index = self.tree.indexAt(pos)
        if index.isValid():
            fileInfo = self.model.fileInfo(index)
            if fileInfo.isFile():
                self.menu.exec(self.mapToGlobal(pos))

    def onMenuRemove(self):
        index = self.tree.currentIndex()
        if index.isValid():
            if self.mw.pm.getPlayer(self.model.filePath(index)):
                QMessageBox.warning(self, "Warning",
                                        "Camera is currently playing. Please stop before deleting.",
                                        QMessageBox.StandardButton.Ok)
                return
            
            ret = QMessageBox.warning(self, "Cayenue",
                                        "You are about to ** PERMANENTLY ** delete this file.\n"
                                        "Are you sure you want to continue?",
                                        QMessageBox.StandardButton.Ok | QMessageBox.StandardButton.Cancel)

            if ret == QMessageBox.StandardButton.Ok:
                try:
                    idxAbove = self.tree.indexAbove(index)
                    idxBelow = self.tree.indexBelow(index)

                    self.model.remove(index)
                    
                    resolved = False
                    if idxAbove.isValid():
                        if os.path.isfile(self.model.filePath(idxAbove)):
                            self.tree.setCurrentIndex(idxAbove)
                            resolved = True
                    if not resolved:
                        if idxBelow.isValid():
                            if os.path.isfile(self.model.filePath(idxBelow)):
                                self.tree.setCurrentIndex(idxBelow)
                
                except Exception as e:
                    logger.error(f'File delete error: {e}')

    def onMenuRename(self):
        player = self.mw.pm.getPlayer(self.mw.filePanel.getCurrentFileURI())
        if player:
            self.mw.onError("Please stop the file playing in order to rename")
            return
        index = self.tree.currentIndex()
        if index.isValid():
            self.model.setReadOnly(False)
            self.tree.edit(index)

    def onFileRenamed(self, path, oldName, newName):
        self.model.setReadOnly(True)

    def onMenuInfo(self):
        strInfo = ""
        try:
            index = self.tree.currentIndex()
            if (index.isValid()):
                info = self.model.fileInfo(index)
                strInfo += "Filename: " + info.fileName()
                strInfo += "\nCreated: " + info.birthTime().toString()
                strInfo += "\nModified: " + info.lastModified().toString()

                reader = avio.Reader(info.absoluteFilePath())
                duration = reader.duration()
                time_in_seconds = int(duration / 1000)
                hours = int(time_in_seconds / 3600)
                minutes = int((time_in_seconds - (hours * 3600)) / 60)
                seconds = int((time_in_seconds - (hours * 3600) - (minutes * 60)))
                strInfo += "\nDuration: " + str(minutes) + ":" + "{:02d}".format(seconds)
                #title = reader.metadata("title")
                #if len(title):
                #    strInfo += "\nTitle: " + reader.metadata("title")

                if (reader.has_video()):
                    strInfo += "\n\nVideo Stream:"
                    strInfo += "\n    Resolution:  " + str(reader.width()) + " x " + str(reader.height())
                    strInfo += "\n    Frame Rate:  " + f'{reader.frame_rate().num / reader.frame_rate().den:.2f}'
                    strInfo += "  (" + str(reader.frame_rate().num) + " / " + str(reader.frame_rate().den) +")"
                    strInfo += "\n    Time Base:  " + str(reader.video_time_base().num) + " / " + str(reader.video_time_base().den)
                    strInfo += "\n    Video Codec:  " + reader.str_video_codec()
                    strInfo += "\n    Pixel Format:  " + reader.str_pix_fmt()
                    strInfo += "\n    Bitrate:  " + f'{reader.video_bit_rate():,}'
                
                if (reader.has_audio()):
                    strInfo += "\n\nAudio Stream:"
                    strInfo += "\n    Channel Layout:  " + reader.str_channel_layout()
                    strInfo += "\n    Audio Codec:  " + reader.str_audio_codec()
                    strInfo += "\n    Sample Rate:  " + str(reader.sample_rate())
                    strInfo += "\n    Sample Size:  " + str(reader.frame_size())
                    strInfo += "\n    Time Base:  " + str(reader.audio_time_base().num) + " / " + str(reader.audio_time_base().den)
                    strInfo += "\n    Sample Format:  " + reader.str_sample_format()
                    strInfo += "\n    Bitrate:  " + f'{reader.audio_bit_rate():,}'
                
            else:
                strInfo = "Invalid Index"
        except Exception as ex:
            strInfo = f'Unable to read file info: {ex}'

        #msgBox = QMessageBox(self)
        #msgBox.setWindowTitle("File Info")
        #msgBox.setText(strInfo)
        #msgBox.exec()
        self.dlgInfo.lblMessage.setText(strInfo)
        self.dlgInfo.exec()

    def onMenuPlay(self):
        self.mw.filePanel.control.btnPlayClicked()

    def onMenuStop(self):
        self.mw.filePanel.control.btnStopClicked()

    def fastForward(self):
        try:
            pct = self.progress.sldProgress.value() / 1000
            if duration := self.progress.duration:
                interval = 10000 / duration
                tgt = pct + interval
                if tgt < 1.0:
                    if player := self.getCurrentlyPlayingFile():
                        player.seek(tgt)
        except Exception as ex:
            logger.error(f"FilePanel fastForward exception: {ex}")

    def rewind(self):
        try:
            pct = self.progress.sldProgress.value() / 1000
            if duration := self.progress.duration:
                interval = 10000 / duration
                tgt = max(pct - interval, 0.0)
                if player := self.getCurrentlyPlayingFile():
                    player.seek(tgt)
        except Exception as ex:
            logger.error(f"FilePanel rewind exception: {ex}")

    def getCurrentFileURI(self):
        result = None
        index = self.tree.currentIndex()
        if index.isValid():
            info = self.model.fileInfo(index)
            if info.isFile():
                result = info.absoluteFilePath()
        return result
    
    def getCurrentlyPlayingFile(self):
        result = None
        self.mw.pm.lock()
        for player in self.mw.pm.players.values():
            if not player.isCameraStream():
                result = player
        self.mw.pm.unlock()
        return result
            
    def setCurrentFile(self, uri):
        index = self.model.index(uri)
        self.tree.setCurrentIndex(index)
        self.control.setBtnPlay()
        self.control.setBtnMute()
        self.control.setSldVolume()
        if player := self.mw.pm.getPlayer(uri):
            self.onMediaProgress(player.file_progress, uri)

    def showEvent(self, event):
        self.restoreHeader()

    def headerChanged(self, a, b, c):
        key = f'File/Header'
        self.mw.settings.setValue(key, self.tree.header().saveState())

    def restoreHeader(self):
        key = f'File/Header'
        data = self.mw.settings.value(key)
        if data:
            self.tree.header().restoreState(data)
        self.tree.update()
        self.tree.header().sectionResized.connect(self.headerChanged)
        self.tree.header().sectionMoved.connect(self.headerChanged)

    def getDirectory(self):
        try:
            if sys.platform == "win32":
                path = os.environ["HOMEPATH"]
            else:
                path = os.environ["HOME"]
            key = f'File/Directory'
            dirs = QStandardPaths.standardLocations(QStandardPaths.StandardLocation.MoviesLocation)
            path = self.mw.settings.value(key, dirs[0])
            os.makedirs(path, exist_ok=True)
        except Exception as ex:
            logger.error(f'Unable to find Videos directory: {ex}')
        return path

    def setDirectory(self, path):
        key = f'File/Directory'
        self.mw.settings.setValue(key, path)

    def getMute(self):
        key = f'File/Mute'
        return bool(int(self.mw.settings.value(key, 0)))
    
    def setMute(self, state):
        key = f'File/Mute'
        self.mw.settings.setValue(key, int(state))

    def getVolume(self):
        key = f'File/Volume'
        return int(self.mw.settings.value(key, 80))
    
    def setVolume(self, volume):
        key = f'File/Volume'
        self.mw.settings.setValue(key, volume)



#/********************************************************************
# Cayenue/cayenue/panels/file/filesearchdialog.py 
#
# Copyright (c) 2025  Stephen Rhodes
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#*********************************************************************/

import os
from PyQt6.QtWidgets import QGridLayout, QWidget, \
    QLabel, QMessageBox, QAbstractItemView, \
    QDialog, QCalendarWidget, QDialogButtonBox, QComboBox, \
    QAbstractItemView
from PyQt6.QtGui import QBrush
from PyQt6.QtCore import Qt
from loguru import logger
from datetime import datetime
from cayenue.enums import Occurence

FORMAT = "%Y%m%d%H%M%S"

class FileSearchDialog(QDialog):
    def __init__(self, mw):
        super().__init__(mw)
        self.mw = mw
        self.geometryKey = "FileSearchDialog/geometry"
        self.timeFormat = ""
        self.positionInitialized = False
        self.setWindowTitle("File Search")

        self.matching_file = None
        self.closest_before = None
        self.closest_after = None

        if rect := self.mw.settings.value(self.geometryKey):
            if rect.width() and rect.height():
                self.setGeometry(rect)
                self.positionInitialized = True

        self.cameras = QComboBox()
        self.pnlCameras = QWidget()
        lytCamera = QGridLayout(self.pnlCameras)
        lytCamera.addWidget(QLabel("Camera"),  0, 0, 1, 1)
        lytCamera.addWidget(self.cameras,      0, 1, 1, 1)
        lytCamera.setColumnStretch(1, 10)

        self.calendar = QCalendarWidget()
        self.calendar.setVerticalHeaderFormat(QCalendarWidget.VerticalHeaderFormat.NoVerticalHeader)
        self.calendar.setStyleSheet("QTableView{selection-background-color: darkGreen; selection-color: lightGray}")
        format = self.calendar.weekdayTextFormat(Qt.DayOfWeek.Saturday)
        format.setForeground(QBrush(Qt.GlobalColor.white, Qt.BrushStyle.SolidPattern))
        self.calendar.setWeekdayTextFormat(Qt.DayOfWeek.Saturday, format)
        self.calendar.setWeekdayTextFormat(Qt.DayOfWeek.Sunday, format)
        
        self.hour = QComboBox()
        self.hour.setMinimumWidth(50)
        hours = []
        for i in range(1, 13):
            hours.append(str(i))
        self.hour.addItems(hours)
        self.hour.setCurrentText(datetime.now().strftime("%I").lstrip('0'))

        self.minute = QComboBox()
        self.minute.setMinimumWidth(50)
        minutes = []
        for i in range(0, 60):
            if i < 10:
                minutes.append(f'0{i}')
            else:
                minutes.append(str(i))
        self.minute.addItems(minutes)
        self.minute.setCurrentText(datetime.now().strftime("%M"))

        self.AM_PM = QComboBox()
        self.AM_PM.setMinimumWidth(60)
        self.AM_PM.addItems(["AM", "PM"])
        self.AM_PM.setCurrentText(datetime.now().strftime("%p"))

        self.pnlTime = QWidget()
        lytTime = QGridLayout(self.pnlTime)
        lytTime.addWidget(QLabel(),        0, 0, 1, 1)
        lytTime.addWidget(self.hour,       0, 1, 1, 1)
        lytTime.addWidget(QLabel(" : "),   0, 2, 1, 1)
        lytTime.addWidget(self.minute,     0, 3, 1, 1)
        lytTime.addWidget(self.AM_PM,      0, 4, 1, 1)
        lytTime.addWidget(QLabel(),        0, 5, 1, 1)
        lytTime.setColumnStretch(0, 2)
        lytTime.setColumnStretch(0, 5)
        lytTime.setColumnStretch(0, 0)
        lytTime.setColumnStretch(0, 5)
        lytTime.setColumnStretch(0, 2)

        self.buttonBox = QDialogButtonBox(QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel)
        self.buttonBox.button(QDialogButtonBox.StandardButton.Cancel).setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.buttonBox.button(QDialogButtonBox.StandardButton.Ok).setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.buttonBox.rejected.connect(self.reject)
        self.buttonBox.accepted.connect(self.accept)

        instruct = QLabel("Select a Camera, Date and Time")

        lytMain = QGridLayout(self)
        lytMain.addWidget(instruct,           0, 0, 1, 1, Qt.AlignmentFlag.AlignCenter)
        lytMain.addWidget(self.pnlCameras,    1, 0, 1, 1)
        lytMain.addWidget(self.calendar,      2, 0, 1, 1)
        lytMain.addWidget(self.pnlTime,       3, 0, 1, 1)
        lytMain.addWidget(self.buttonBox,     4, 0, 1, 1)

    def moveEvent(self, event):
        # for adding persistence in the future maybe
        return super().moveEvent(event)

    def resizeEvent(self, event):
        # for adding persistence in the future maybe
        return super().resizeEvent(event)

    def reject(self):
        self.matching_file = None
        self.closest_before = None
        self.closest_after = None
        self.hide()

    def accept(self):
        self.matching_file = None
        self.closest_before = None
        self.closest_after = None
        try:
            selected = self.getSelectedDate()
            main_directory = self.mw.filePanel.dirArchive.txtDirectory.text()
            sub_directory = self.cameras.currentText()
            self.findFileForEventTime(selected, main_directory,  sub_directory)

            if self.matching_file:
                self.selectFileInTree(os.path.join(main_directory, sub_directory), self.matching_file)
            self.hide()

            if self.matching_file:
                answer = QMessageBox.question(self.mw, "Found Event Time", "The program found a match, would you like to start the playback?")
                if answer == QMessageBox.StandardButton.Yes:
                    main_directory = self.mw.filePanel.dirArchive.txtDirectory.text()
                    sub_directory = self.cameras.currentText()
                    tree = self.mw.filePanel.tree
                    model = tree.model()
                    path = os.path.join(main_directory, sub_directory, self.matching_file)
                    if file_idx := model.index(path):
                        if file_idx.isValid():
                            # repeat select file for first pass bug
                            self.selectFileInTree(os.path.join(main_directory, sub_directory), self.matching_file)
                            
                            file_start_time = self.fileAsDate(self.matching_file).timestamp()
                            file_end_time = self.endTimestamp(os.path.join(main_directory, sub_directory), self.matching_file)
                            file_seek_time = selected.timestamp()
                            pct = 0
                            num = file_seek_time - file_start_time
                            den = file_end_time - file_start_time
                            if den:
                                pct = num/den
                            if pct > .99:
                                pct = 0
                            self.mw.filePanel.control.startPlayer(file_start_from_seek=pct)
            else:
                file_to_index = None
                dist_to_before = None
                dist_to_after = None
                if self.closest_before:
                    dist_to_before = selected.timestamp() - self.endTimestamp(os.path.join(main_directory, sub_directory), self.closest_before)

                if self.closest_after:
                    dist_to_after = self.startTimestamp(self.closest_after) - selected.timestamp()

                found = False
                if dist_to_before and not dist_to_after:
                    found = True
                    file_to_index = self.closest_before
                if not dist_to_before and dist_to_after:
                    found = True
                    file_to_index = self.closest_after
                if dist_to_before and dist_to_after:
                    found = True
                    if dist_to_before < dist_to_after:
                        file_to_index = self.closest_before
                    else:
                        file_to_index = self.closest_after

                if not found:
                    QMessageBox.warning(self.mw, "Algorithm Error", "The program was not able to resolve the file")
                else:
                    self.selectFileInTree(os.path.join(main_directory, sub_directory), file_to_index)
                    answer = QMessageBox.question(self.mw, "Closest Result", "The program could not find an exact match, the closest result is highlighted, would you like to open it?")
                    self.selectFileInTree(os.path.join(main_directory, sub_directory), file_to_index)
                    if answer == QMessageBox.StandardButton.Yes:
                        self.mw.filePanel.control.startPlayer()

        except Exception as ex:
            logger.error(f'File search error: {ex}')

        self.hide()

    def selectFileInTree(self, path, filename):
        tree = self.mw.filePanel.tree
        model = tree.model()
        if camera_idx := model.index(path):
            if camera_idx.isValid():
                if not tree.isExpanded(camera_idx):
                    tree.setExpanded(camera_idx, True)
                if file_idx := model.index(os.path.join(path, filename)):
                    if file_idx.isValid():
                        tree.setCurrentIndex(file_idx)
                        tree.scrollTo(file_idx, QAbstractItemView.ScrollHint.PositionAtCenter)

    def qualifiedFileName(self, path, name):
        if not os.path.isfile(os.path.join(path, name)):
            return False
        components = os.path.splitext(name)
        if len(components) != 2:
            return False
        if components[1] != ".mov" and components[1] != ".mp4":
            return False
        if not components[0].isdigit():
            return False
        if len(components[0]) != 14:
            return False
        return True

    def isBefore(self, target, filename):
        result = False
        reference = datetime.strptime(os.path.splitext(filename)[0], FORMAT)
        reference = reference.replace(second=0)
        if target <= reference:
            result = True
        return result
    
    def isAfter(self, target, path, filename):
        result = False
        reference = datetime.fromtimestamp(os.path.getmtime(os.path.join(path, filename)))
        reference = reference.replace(second=59)
        if target >= reference:
            result = True
        return result
    
    def startTimestamp(self, filename):
        # file start time is deduced from file name
        return datetime.strptime(os.path.splitext(filename)[0], FORMAT).timestamp()
    
    def endTimestamp(self, path, filename):
        # file end time is the os file creation time
        return datetime.fromtimestamp(os.path.getmtime(os.path.join(path, filename))).timestamp()
    
    def fileAsDate(self, file):
        return datetime.strptime(os.path.splitext(file)[0], FORMAT)

    def getOccurence(self, target, path, filename):
        result = Occurence.DURING
        if self.isBefore(target, filename):
            result = Occurence.BEFORE
        if self.isAfter(target, path, filename):
            result = Occurence.AFTER
        return result
    
    def getSelectedDate(self):
        result = None
        date = self.calendar.selectedDate()
        h = int(self.hour.currentText())
        if self.AM_PM.currentText() == "PM" and h < 12:
            h += 12
        if self.AM_PM.currentText() == "AM" and h == 12:
            h = 0
        tmp = f'{date.year()}{date.month():02}{date.day():02}{h:02}{self.minute.currentText()}30'
        result = datetime.strptime(tmp, FORMAT)
        return result

    def guessFileIndex(self, selected, path, files, max_idx, min_idx, last_idx):
        idx = int(min_idx + (max_idx - min_idx) / 2)
  
        if idx == last_idx:
            # algorithm converged without finding match
            if self.getOccurence(selected, path, files[idx]) == Occurence.BEFORE:
                if idx > 0:
                    self.closest_before = files[idx-1]
                self.closest_after = files[idx]

            if self.getOccurence(selected, path, files[idx]) == Occurence.AFTER:
                self.closest_before = files[idx]
                if idx < len(files) - 1:
                    self.closest_after = files[idx+1]
            return

        last_idx = idx
        
        match self.getOccurence(selected, path, files[idx]):
            case Occurence.BEFORE:
                self.guessFileIndex(selected, path, files, idx, min_idx, last_idx)
            case Occurence.AFTER:
                self.guessFileIndex(selected, path, files, max_idx, idx, last_idx)
            case Occurence.DURING:
                self.matching_file = files[idx]
        return
    
    def findFileForEventTime(self, target_time, main_directory, sub_directory):
        path = os.path.join(main_directory, sub_directory)
        files = os.listdir(path)
        files = [f for f in files if self.qualifiedFileName(path, f)]
        files.sort()
    
        self.matching_file = None
        self.closest_before = None
        self.closest_after = None

        inside_range = True
        if self.isBefore(target_time, files[0]):
            inside_range = False
            self.closest_after = files[0]
        if self.isAfter(target_time, path, files[-1]):
            inside_range = False
            self.closest_before = files[-1]
        if inside_range:
            self.guessFileIndex(target_time, path, files, len(files)-1, 0, -1)




#********************************************************************
# Cayenue/cayenue/protocols/server.py
#
# Copyright (c) 2025  Stephen Rhodes
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#*********************************************************************/

import libonvif as onvif
from loguru import logger
import numpy as np
import io
from cayenue.panels.camera import Snapshot

class ServerProtocols():
    def __init__(self, mw):
        self.mw = mw
        self.snapshot = Snapshot(mw)

    def callback(self, msg):
        buffer = io.BytesIO()
        args = msg.split("\n\n")

        match args[0]:
            case "GET CAMERAS":
                lstCamera = self.mw.cameraPanel.lstCamera
                cameras = [lstCamera.item(x) for x in range(lstCamera.count())]
                buffer.write(bytearray("GET CAMERAS\n\n", 'utf-8'))
                for c_idx, camera in enumerate(cameras):
                    for p_idx, profile in enumerate(camera.profiles):
                        buffer.write(bytearray(profile.toJSON(), 'utf-8'))
                        if c_idx < len(cameras) - 1 :
                            buffer.write(bytearray("\n", 'utf-8'))
                        else:
                            if p_idx < len(camera.profiles) - 1:
                                buffer.write(bytearray("\n", 'utf-8'))
                    if c_idx < len(cameras) - 1:
                        buffer.write(bytearray("\n", 'utf-8'))
                buffer.write(bytearray("\r\n", 'utf-8'))

            case "UPDATE VIDEO":
                data = onvif.Data(args[1])
                data.updateVideo()
                buffer.write(self.resolve(data))

            case "UPDATE AUDIO":
                data = onvif.Data(args[1])
                data.updateAudio()
                buffer.write(self.resolve(data))

            case "UPDATE IMAGE":
                data = onvif.Data(args[1])
                data.updateImage()
                buffer.write(self.resolve(data))

            case "MOVE":
                data = onvif.Data(args[1])
                data.move()
                buffer.write(bytearray("PTZ\r\n", 'utf-8'))

            case "STOP":
                data = onvif.Data(args[1])
                data.stop()
                buffer.write(bytearray("PTZ\r\n", 'utf-8'))

            case "GOTO PRESET":
                data = onvif.Data(args[1])
                data.set()
                buffer.write(bytearray("GOTO PRESET\r\n", 'utf-8'))

            case "SET PRESET":
                data = onvif.Data(args[1])
                data.setGotoPreset()
                buffer.write(bytearray("SET PRESET\r\n", 'utf-8'))

            case "REBOOT":
                data = onvif.Data(args[1])
                data.reboot()
                buffer.write(bytearray("REBOOT\r\n", 'utf-8'))

            case "SYNC TIME":
                data = onvif.Data(args[1])
                if camera := self.mw.cameraPanel.getCameraBySerialNumber(data.serial_number()):
                    camera.onvif_data.updateTime()
                buffer.write(bytearray("SYNC TIME\r\n", 'utf-8'))

            case "SNAPSHOT":
                profile = onvif.Data(args[1])
                remote = self.mw.settingsPanel.proxy.lblServer.text().split()[0].strip()
                key = f'{remote}{profile.serial_number()}/{profile.profile()}'
                if camera := self.mw.cameraPanel.getCamera(key):
                    buffer.write(bytearray("SNAPSHOT\n\n", 'utf-8'))
                    buffer.write(bytearray(profile.user_data() + "\r\n", 'utf-8'))
                    self.snapshot.getBufferedSnapshot(profile, buffer, camera)

        return np.frombuffer(buffer.getvalue(), dtype=np.uint8)
    
    def resolve(self, data):
        if camera := self.mw.cameraPanel.getCameraBySerialNumber(data.serial_number()):
            camera.syncData(data)
        return bytearray("UPDATE\n\n" + data.toJSON() + "\r\n", 'utf-8')

    def error(self, msg):
        logger.error(f"server protocol error: {msg}")


#/********************************************************************
# Cayenue/cayenue/panels/settings/general.py 
#
# Copyright (c) 2024  Stephen Rhodes
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#*********************************************************************/

import os
import sys
from PyQt6.QtWidgets import QMessageBox, QLineEdit, QSpinBox, \
    QGridLayout, QWidget, QCheckBox, QLabel, QComboBox, QPushButton, \
    QDialog, QTextEdit, QMessageBox, QFileDialog, QGroupBox, \
    QListWidget, QDialogButtonBox, QListWidgetItem
from PyQt6.QtCore import QFile, QRect, Qt, QSettings
from PyQt6.QtGui import QTextCursor, QTextOption
from loguru import logger
import webbrowser
import platform
from cayenue.player import Player
from cayenue.enums import Style, ProxyType
import cayenue
from pathlib import Path
import subprocess

class LogText(QTextEdit):
    def __init__(self, parent):
        super().__init__(parent)
        self.setWordWrapMode(QTextOption.WrapMode.NoWrap)

    def scrollToBottom(self):
        self.moveCursor(QTextCursor.MoveOperation.End, QTextCursor.MoveMode.MoveAnchor)
        self.ensureCursorVisible()

class LogDialog(QDialog):
    def __init__(self, mw):
        super().__init__(mw)
        self.mw = mw
        self.geometryKey = "LogDialog/geometry"
        if rect := self.mw.settings.value(self.geometryKey):
            if rect.width() and rect.height():
                self.setGeometry(rect)
        else:
            g = mw.geometry()
            rect = QRect(0, 0, g.width()-20, 480)
            rect.moveCenter(g.center())
            self.setGeometry(rect)

        self.editor = LogText(self)
        self.editor.setReadOnly(True)
        self.editor.setFontFamily("Courier")

        self.lblSize = QLabel("Log File Size:")
        self.btnArchive = QPushButton("View Archive")
        self.btnArchive.clicked.connect(self.btnArchiveClicked)
        self.btnClear = QPushButton("  Clear  ")
        self.btnClear.clicked.connect(self.btnClearClicked)
        self.btnClose = QPushButton("Close")
        self.btnClose.clicked.connect(self.btnCloseClicked)

        pnlButton = QWidget()
        lytButton = QGridLayout(pnlButton)
        lytButton.addWidget(self.btnArchive,   0, 1, 1, 1)
        lytButton.addWidget(self.btnClear,     0, 2, 1, 1)
        lytButton.addWidget(QLabel(),          0, 3, 1, 1)
        lytButton.addWidget(self.btnClose,     0, 4, 1, 1, Qt.AlignmentFlag.AlignRight)
        lytButton.setContentsMargins(0, 0, 0, 0)
        lytButton.setColumnStretch(3, 10)
        
        panel = QWidget()
        lytPanel = QGridLayout(panel)
        lytPanel.addWidget(self.lblSize,    0, 0, 1, 1)
        lytPanel.addWidget(pnlButton,       0, 1, 1, 1)

        lyt = QGridLayout(self)
        lyt.addWidget(self.editor, 0, 0, 1, 1)
        lyt.addWidget(panel,       1, 0, 1, 1)
        lyt.setRowStretch(0, 10)

    def closeEvent(self, e):
        self.mw.settings.setValue(self.geometryKey, self.geometry())

    def readLogFile(self, path):
        data = ""
        if os.path.isfile(path):
            with open(path, 'r') as file:
                data = file.read()
        self.setWindowTitle(path)
        self.editor.setText(data)
        y = "{:.2f}".format(os.stat(path).st_size/1000000)
        self.lblSize.setText(f'Log File Size: {y}  MB    ')
        self.editor.scrollToBottom()

    def btnArchiveClicked(self):
        path = None
        if platform.system() == "Linux":
            path = QFileDialog.getOpenFileName(self, "Select File", self.windowTitle(), options=QFileDialog.Option.DontUseNativeDialog)[0]
        else:
            path = QFileDialog.getOpenFileName(self, "Select File", self.windowTitle())[0]
        if path:
            if len(path) > 0:
                self.readLogFile(os.path.realpath(path))

    def btnCloseClicked(self):
        self.close()

    def btnClearClicked(self):
        filename = self.windowTitle()
        log_filename = self.mw.getLogFilename()

        if filename == log_filename:
            ret = QMessageBox.question(self, "Preserve File", 
                                    "Would you like to preserve the current log file with a new name?",
                                    QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No)
            
            if ret == QMessageBox.StandardButton.Yes:
                path = None
                if platform.system() == "Linux":
                    path = QFileDialog.getSaveFileName(self, "Select A New File Name", self.windowTitle(), "Text (*.txt)", options=QFileDialog.Option.DontUseNativeDialog)[0]
                else:
                    path = QFileDialog.getSaveFileName(self, "Select A New File Name", self.windowTitle(), "Text (*.txt)")[0]
                if path:
                    if len(path) > 0:
                        try:
                            logger.remove(self.mw.logger_id)
                            os.rename(filename, path)
                        except Exception as ex:
                            QMessageBox.warning(self, "File Error", f'File rename error occurred {ex}')
                        self.mw.logger_id = logger.add(log_filename)
                        logger.debug("Created new log file")
                        self.editor.setText("")
                        self.readLogFile(log_filename)

            else:
                ret = QMessageBox.warning(self, "Deleting Current Log",
                                            "You are about to delete the current log.\n"
                                            "Are you sure you want to continue?",
                                            QMessageBox.StandardButton.Ok | QMessageBox.StandardButton.Cancel)
                if ret == QMessageBox.StandardButton.Ok:
                    try:
                        logger.remove(self.mw.logger_id)
                        os.remove(filename)
                        self.editor.setText("")
                    except Exception as ex:
                        QMessageBox.warning(self, "File Error", f'File delete error occurred {ex}')

                    self.logger_id = logger.add(log_filename)
                    logger.debug("Created new log file")
                    self.readLogFile(log_filename)

        else:
            ret = QMessageBox.warning(self, "Deleting File",
                                        f'\n{filename}\n\n'
                                        "You are about to delete this file.\n"
                                        "Are you sure you want to continue?\n\n"
                                        "Clicking cancel will preserve the file and return you to the current log",
                                        QMessageBox.StandardButton.Ok | QMessageBox.StandardButton.Cancel)
            if ret == QMessageBox.StandardButton.Ok:
                QFile.remove(filename)
            self.readLogFile(log_filename)
            QMessageBox.information(self, "Current Log Displayed", "The current log has been loaded into the display", QMessageBox.StandardButton.Ok)


class ProfileItem(QListWidgetItem):
    def __init__(self, name):
        super().__init__(name)
        self.original = name
        if name != "Focus" and name != "Reader":
            self.setFlags(self.flags() | Qt.ItemFlag.ItemIsEditable)

class ProfileDialog(QDialog):
    def __init__(self, mw):
        super().__init__(mw)
        self.mw = mw
        self.editingItem = None
        self.deleted = []

        self.setModal(True)
        self.setWindowTitle("Manage Profiles")
        self.list = QListWidget()

        self.list.itemDoubleClicked.connect(self.onItemDoubleClicked)
        self.list.itemSelectionChanged.connect(self.onItemSelectionChanged)

        self.buttonBox = QDialogButtonBox(QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel)
        self.buttonBox.button(QDialogButtonBox.StandardButton.Cancel).setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.buttonBox.rejected.connect(self.reject)
        self.buttonBox.accepted.connect(self.accept)

        self.btnAdd = QPushButton("+")
        self.btnAdd.clicked.connect(self.btnAddClicked)
        self.btnDelete = QPushButton(" - ")
        self.btnDelete.clicked.connect(self.btnDeleteClicked)
        self.btnUp = QPushButton("^")
        self.btnUp.clicked.connect(self.btnUpClicked)
        self.btnDown = QPushButton("v")
        self.btnDown.clicked.connect(self.btnDownClicked)

        lytMain = QGridLayout(self)
        lytMain.addWidget(QLabel(),         0, 0, 1, 1)
        lytMain.addWidget(self.btnUp,      1, 0, 1, 1, Qt.AlignmentFlag.AlignCenter)
        lytMain.addWidget(self.btnDown,    2, 0, 1, 1, Qt.AlignmentFlag.AlignCenter)
        lytMain.addWidget(QLabel(),        3, 0, 1, 1)
        lytMain.addWidget(self.list,       0, 1, 4, 5)
        lytMain.addWidget(self.btnAdd,     4, 1, 1, 1, Qt.AlignmentFlag.AlignRight)
        lytMain.addWidget(self.btnDelete,  4, 2, 1, 1, Qt.AlignmentFlag.AlignLeft)
        lytMain.addWidget(self.buttonBox,  4, 3, 1, 3)
        lytMain.setRowStretch(0, 10)
        lytMain.setRowStretch(3, 10)

    def reject(self):
        self.hide()

    def accept(self):
        self.clearProfileNames()
        self.mw.settingsPanel.general.cmbViewerProfile.clear()
        for i in range(self.list.count()):
            name = self.list.item(i).text() 
            key = f'SettingsProfile_{i}'
            self.mw.settings.setValue(key, name)
            self.mw.settingsPanel.general.cmbViewerProfile.addItem(name)
        for name in self.deleted:
            settings = QSettings("onvif", name.text())
            settings.clear()
        self.hide()

    def onItemDoubleClicked(self, item):
        self.editingItem = item
        self.list.editItem(item)
        
    def onItemSelectionChanged(self):
        self.syncGui()

    def btnAddClicked(self):
        row = self.list.currentRow()+1
        item = ProfileItem("New Profile")
        self.list.insertItem(row, item)
        self.list.setCurrentRow(row)
        self.list.itemDoubleClicked.emit(item)
        self.syncGui()

    def btnDeleteClicked(self):
        self.deleted.append(self.list.takeItem(self.list.currentRow()))
        self.syncGui()

    def btnUpClicked(self):
        row = self.list.currentRow()
        item = self.list.takeItem(row)
        self.list.insertItem(row-1, item)
        self.list.setCurrentRow(row-1)
        self.syncGui()

    def btnDownClicked(self):
        row = self.list.currentRow()
        item = self.list.takeItem(row)
        self.list.insertItem(row+1, item)
        self.list.setCurrentRow(row+1)
        self.syncGui()

    def syncGui(self):
        self.btnDelete.setEnabled(self.list.count() > 1)
        if self.list.count():
            self.btnUp.setEnabled(self.list.currentRow() > 0)
            self.btnDown.setEnabled(self.list.currentRow() < self.list.count()-1)
        if item := self.list.currentItem():
            if item.text() == "Focus" or item.text() == "Reader":
                self.btnDelete.setEnabled(False)

    def exec(self):
        self.list.clear()
        self.deleted.clear()
        cmb = self.mw.settingsPanel.general.cmbViewerProfile
        for i in range(cmb.count()):
            self.list.addItem(ProfileItem(cmb.itemText(i)))
        self.list.setCurrentRow(0)
        self.list.setFocus()
        self.syncGui()
        super().exec()

    def clearProfileNames(self):
        i = 0
        while self.mw.settings.contains(f'SettingsProfile_{i}'):
            self.mw.settings.remove(f'SettingsProfile_{i}')
            i += 1

class GeneralOptions(QWidget):
    def __init__(self, mw):
        super().__init__()
        self.mw = mw
        self.process = None
        self.thread = None

        self.usernameKey = "settings/username"
        self.passwordKey = "settings/password"
        self.decoderKey = "settings/decoder"
        self.bufferSizeKey = "settings/bufferSize"
        self.lagTimeKey = "settings/lagTime"
        self.startFullScreenKey = "settings/startFullScreen"
        self.autoTimeSyncKey = "settings/autoTimeSync"
        self.alarmSoundFileKey = "settings/alarmSoundFile"
        self.alarmSoundVolumeKey = "settings/alarmSoundVolume"
        self.displayRefreshKey = "settings/displayRefresh"
        self.audioDriverIndexKey = "settings/audioDriverIndex"
        self.appearanceKey = "settings/appearance"
        self.displaySizeKey = "settings/displaySize"
        self.stressTestKey = "settings/stressTest"
        self.snapshotDlgKey = "settings/snapshotDlg"

        p = Player("", self)
        audioDrivers = p.getAudioDrivers()

        decoders = ["none"] + p.getHardwareDecoders()

        self.mediamtx_process = None

        self.dlgLog = None
        self.dlgProfiles = None

        self.txtUsername = QLineEdit()
        self.txtUsername.setText(mw.settings.value(self.usernameKey, ""))
        self.txtUsername.textChanged.connect(self.usernameChanged)
        lblUsername = QLabel("Common Username")
        
        self.txtPassword = QLineEdit()
        self.txtPassword.setText(mw.settings.value(self.passwordKey, ""))
        self.txtPassword.textChanged.connect(self.passwordChanged)
        lblPassword = QLabel("Common Password")
        
        self.cmbDecoder = QComboBox()
        self.cmbDecoder.addItems(decoders)
        self.cmbDecoder.setCurrentText(mw.settings.value(self.decoderKey, "none"))
        self.cmbDecoder.currentTextChanged.connect(self.cmbDecoderChanged)
        lblDecoders = QLabel("Hardware Decoder")

        self.cmbAudioDriver = QComboBox()
        self.cmbAudioDriver.addItems(audioDrivers)
        self.cmbAudioDriver.setCurrentIndex(int(mw.settings.value(self.audioDriverIndexKey, 0)))
        self.cmbAudioDriver.currentIndexChanged.connect(self.cmbAudioDriverChanged)
        lblAudioDrivers = QLabel("Audio Driver")

        self.cmbAppearance = QComboBox()
        self.cmbAppearance.addItems(["Dark", "Light"])
        self.cmbAppearance.setCurrentText(mw.settings.value(self.appearanceKey, "Dark"))
        self.cmbAppearance.currentTextChanged.connect(self.cmbAppearanceChanged)
        lblAppearance = QLabel("Appearance")

        self.chkStartFullScreen = QCheckBox("Start Full Screen")
        self.chkStartFullScreen.setChecked(bool(int(mw.settings.value(self.startFullScreenKey, 0))))
        self.chkStartFullScreen.stateChanged.connect(self.startFullScreenChecked)

        self.chkAutoTimeSync = QCheckBox("Auto Time Sync")
        self.chkAutoTimeSync.setChecked(bool(int(mw.settings.value(self.autoTimeSyncKey, 0))))
        self.chkAutoTimeSync.stateChanged.connect(self.autoTimeSyncChecked)

        self.chkStressTest = QCheckBox("Stress Test Cycle")
        self.chkStressTest.setChecked(False)
        self.chkStressTest.stateChanged.connect(self.stressTestChecked)

        self.chkSnapshotDlg = QCheckBox("Snapshot File Dlg")
        self.chkSnapshotDlg.setChecked(bool(int(mw.settings.value(self.snapshotDlgKey, 1))))
        self.chkSnapshotDlg.stateChanged.connect(self.snapshotDlgChecked)

        pnlChecks = QWidget()
        lytChecks = QGridLayout(pnlChecks)
        lytChecks.addWidget(self.chkStartFullScreen,  0, 0, 1, 1)
        lytChecks.addWidget(self.chkAutoTimeSync,     0, 1, 1, 1)
        #lytChecks.addWidget(self.chkStressTest,       1, 0, 1, 1)
        lytChecks.addWidget(self.chkSnapshotDlg,      1, 1, 1, 1)

        self.spnDisplayRefresh = QSpinBox()
        self.spnDisplayRefresh.setMinimum(1)
        self.spnDisplayRefresh.setMaximum(1000)
        self.spnDisplayRefresh.setMaximumWidth(80)
        refresh = 10
        if sys.platform == "win32":
            refresh = 20
        self.spnDisplayRefresh.setValue(int(self.mw.settings.value(self.displayRefreshKey, refresh)))
        self.spnDisplayRefresh.valueChanged.connect(self.spnDisplayRefreshChanged)
        lblDisplayRefresh = QLabel("Display Refresh Interval (ms)")

        self.btnCloseAll = QPushButton("Start All")
        self.btnCloseAll.clicked.connect(self.btnCloseAllClicked)

        self.btnShowLogs = QPushButton("Show Logs")
        self.btnShowLogs.clicked.connect(self.btnShowLogsClicked)

        self.btnHideDisplay = QPushButton("Hide Display")
        self.btnHideDisplay.clicked.connect(self.btnHideDisplayClicked)

        self.btnClearSettings = QPushButton("Clear All Settings")
        self.btnClearSettings.clicked.connect(self.btnClearSettingsClicked)

        self.btnHelp = QPushButton("Help")
        self.btnHelp.clicked.connect(self.btnHelpClicked)

        self.btnTest = QPushButton("Test")
        self.btnTest.clicked.connect(self.btnTestClicked)

        self.lblVersion = QLabel(f'Cayenue Version {mw.version}')

        grpNewView = QGroupBox("Open New Window")
        self.btnNewView = QPushButton("Open")
        self.btnNewView.clicked.connect(self.btnNewViewClicked)
        self.cmbViewerProfile = QComboBox()

        if self.mw.settings_profile == "gui":
            names = self.getProfileNames()
            for name in names:
                self.cmbViewerProfile.addItem(name)
        else:
            grpNewView.setEnabled(False)
        
        self.btnManageProfiles = QPushButton("...")
        self.btnManageProfiles.clicked.connect(self.btnManageProfilesClicked)
        lytNewView = QGridLayout(grpNewView)
        lytNewView.addWidget(self.btnNewView,        0, 0, 1, 1, Qt.AlignmentFlag.AlignLeft)
        lytNewView.addWidget(QLabel("Profile"),      0, 1, 1, 1)
        lytNewView.addWidget(self.cmbViewerProfile,  0, 2, 1, 1)
        lytNewView.addWidget(self.btnManageProfiles, 0, 3, 1, 1, Qt.AlignmentFlag.AlignRight)
        lytNewView.setColumnStretch(2, 10)

        pnlBuffer = QWidget()
        lytBuffer = QGridLayout(pnlBuffer)
        lytBuffer.addWidget(lblDisplayRefresh,      4, 0, 1, 3)
        lytBuffer.addWidget(self.spnDisplayRefresh, 4, 3, 1, 1)
        lytBuffer.setContentsMargins(0, 0, 0, 0)

        pnlButtons = QWidget()
        lytButtons = QGridLayout(pnlButtons)
        lytButtons.addWidget(self.btnHideDisplay,   0, 0, 1, 1)
        lytButtons.addWidget(self.btnShowLogs,      0, 1, 1, 1)
        lytButtons.addWidget(self.btnHelp,          0, 2, 1, 1)
        #lytButtons.addWidget(self.btnCloseAll,      0, 0, 1, 1)
        lytButtons.addWidget(self.btnClearSettings, 1, 0, 1, 3)
        #lytButtons.addWidget(self.btnTest,   1, 1, 1, 1)

        self.lblMemory = QLabel()

        lytMain = QGridLayout(self)
        lytMain.addWidget(lblUsername,         1, 0, 1, 1)
        lytMain.addWidget(self.txtUsername,    1, 1, 1, 1)
        lytMain.addWidget(lblPassword,         2, 0, 1, 1)
        lytMain.addWidget(self.txtPassword,    2, 1, 1, 1)
        lytMain.addWidget(lblDecoders,         3, 0, 1, 1)
        lytMain.addWidget(self.cmbDecoder,     3, 1, 1, 1)
        lytMain.addWidget(lblAudioDrivers,     4, 0, 1, 1)
        lytMain.addWidget(self.cmbAudioDriver, 4, 1, 1, 1)
        lytMain.addWidget(lblAppearance,       5, 0, 1, 1)
        lytMain.addWidget(self.cmbAppearance,  5, 1, 1, 1)
        lytMain.addWidget(pnlChecks,           6, 0, 1, 3)
        lytMain.addWidget(pnlBuffer,           7, 0, 1, 3)
        lytMain.addWidget(grpNewView,          8, 0, 1, 3)
        lytMain.addWidget(pnlButtons,          9, 0, 1, 3)
        lytMain.addWidget(self.lblVersion,    10, 0, 1, 3)
        lytMain.addWidget(self.lblMemory,     11, 0, 1, 3)
        lytMain.setRowStretch(11, 10)

    def usernameChanged(self, username):
        self.mw.settings.setValue(self.usernameKey, username)

    def passwordChanged(self, password):
        self.mw.settings.setValue(self.passwordKey, password)

    def cmbDecoderChanged(self, decoder):
        self.mw.settings.setValue(self.decoderKey, decoder)

    def cmbAudioDriverChanged(self, index):
        self.mw.settings.setValue(self.audioDriverIndexKey, index)
        QMessageBox.warning(self.mw, "Application Restart Required", "It is necessary to re-start Onvif GUI in order to enable this change")

    def cmbAppearanceChanged(self, text):
        self.mw.settings.setValue(self.appearanceKey, text)
        if text == "Dark":
            self.mw.setStyleSheet(self.mw.style(Style.DARK))
        if text == "Light":
            self.mw.setStyleSheet(self.mw.style(Style.LIGHT))

    def getProfileNames(self):
        result = []
        i = 0
        while self.mw.settings.contains(f'SettingsProfile_{i}'):
            result.append(self.mw.settings.value(f'SettingsProfile_{i}'))
            i += 1
        if not len(result):
            result.append("Reader")
            result.append("Focus")
        return result

    def autoDiscoverChecked(self, state):
        self.mw.settings.setValue(self.autoDiscoverKey, state)

    def startFullScreenChecked(self, state):
        self.mw.settings.setValue(self.startFullScreenKey, state)

    def autoTimeSyncChecked(self, state):
        self.mw.settings.setValue(self.autoTimeSyncKey, state)
        self.mw.cameraPanel.enableAutoTimeSync(state)
        if state:
            self.mw.cameraPanel.timeSync()

    def stressTestChecked(self, state):
        ...

    def snapshotDlgChecked(self, state):
        self.mw.settings.setValue(self.snapshotDlgKey, state)

    def spnDisplayRefreshChanged(self, i):
        self.mw.settings.setValue(self.displayRefreshKey, i)
        self.mw.glWidget.timer.setInterval(i)

    def cmbInterfacesChanged(self, network):
        self.mw.settings.setValue(self.interfaceKey, network)
     
    def btnCloseAllClicked(self):
        if self.btnCloseAll.text() == "Close All":
            self.mw.closeAllStreams()
        else:
            self.mw.startAllCameras()
            self.btnCloseAll.setText("Close All")

    def btnShowLogsClicked(self):
        filename = self.mw.getLogFilename()
        if not self.dlgLog:
            self.dlgLog = LogDialog(self.mw)
        self.dlgLog.readLogFile(filename)
        self.dlgLog.setModal(False)
        self.dlgLog.show()

    def btnHelpClicked(self):
        result = webbrowser.get().open("https://github.com/sr99622/libonvif#readme-ov-file")
        if not result:
            webbrowser.get().open("https://github.com/sr99622/libonvif")

    def btnHideDisplayClicked(self):
        controlSize = self.mw.split.sizes()[1]
        if self.mw.split.sizes()[0]:
            self.mw.hide()
            self.mw.settings.setValue(self.displaySizeKey, self.mw.split.sizes()[0])
            rect = self.mw.geometry()
            rect.setX(rect.x() + rect.width() - controlSize)
            self.mw.split.setSizes([0, controlSize])
            self.mw.setGeometry(rect)
            self.mw.show()
            self.btnHideDisplay.setText("Show Display")
        else:
            self.mw.hide()
            displaySize = int(self.mw.settings.value(self.displaySizeKey, 400))
            rect = self.mw.geometry()
            rect.setX(rect.x() - displaySize)
            self.mw.split.setSizes([displaySize, controlSize])
            self.mw.setGeometry(rect)
            self.mw.show()
            self.btnHideDisplay.setText("Hide Display")

    def openFocusWindow(self):
        if not self.mw.focus_window:
            proxy = None
            match self.mw.settingsPanel.proxy.proxyType:
                case ProxyType.CLIENT:
                    proxy = self.mw.settingsPanel.proxy.txtRemote.text()
                case ProxyType.SERVER:
                    proxy = self.mw.settingsPanel.proxy.lblServer.text().split()[0]
            focus_settings = QSettings("Cayenue", "Focus")
            focus_settings.setValue("settings/proxyType", ProxyType.CLIENT)
            focus_settings.setValue("settings/proxyRemote", proxy)
            focus_settings.setValue("settings/autoDiscover", 1)
            #focus_settings.setValue(self.collapsedKey, 1)
            self.mw.focus_window = cayenue.main.MainWindow(settings_profile="Focus")
        self.mw.focus_window.show()

    def btnNewViewClicked(self):
        from cayenue.main import MainWindow
        if self.cmbViewerProfile.currentText() == "Focus":
            if self.mw.settingsPanel.proxy.proxyType == ProxyType.STAND_ALONE:
                QMessageBox.warning(self.mw, "Feature Unavailable", "Focus Window is not available in Stand Alone Configuration, please use Proxy Server mode to enable this feature", QMessageBox.StandardButton.Ok)
                return
            self.mw.openFocusWindow()
        elif self.cmbViewerProfile.currentText() == "Reader":
            reader_settings = QSettings("Cayenue", "Reader")
            reader_settings.setValue("filePanel/hideCameraPanel", 1)
            main_file = Path(__file__).parent.parent.parent / "main.py"
            subprocess.Popen(["python", str(main_file), "--profile", "reader"], env=os.environ.copy(), start_new_session=True, shell=True)
        else:
            window = MainWindow(settings_profile = self.cmbViewerProfile.currentText())
            window.parent_window = self.mw
            self.mw.external_windows.append(window)
            window.show()

    def btnManageProfilesClicked(self):
        if not self.dlgProfiles:
            self.dlgProfiles = ProfileDialog(self.mw)
        self.dlgProfiles.exec()

    def btnClearSettingsClicked(self):
        ret = QMessageBox.warning(self, "Clear Settings", 
                                "This action will clear all program settings, are you sure you want to do this?",
                                QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No)
        if ret == QMessageBox.StandardButton.Yes:
            items = [self.cmbViewerProfile.itemText(i) for i in range(self.cmbViewerProfile.count())]
            for item in items:
                print(item)
                QSettings("Cayenue", item).clear()
                QSettings("Cayenue", "gui").clear()
            QMessageBox.information(self, "Clear Settings",
                                    "All settings have been cleared, please re-start the application",
                                    QMessageBox.StandardButton.Ok)
    
    def btnTestClicked(self):
        print("btnTestClicked")




#/********************************************************************
# Cayenue/cayenue/panels/cameras/datastructures.py 
#
# Copyright (c) 2023  Stephen Rhodes
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#*********************************************************************/

from PyQt6.QtWidgets import QListWidgetItem
from PyQt6.QtCore import Qt, pyqtSignal, QObject, QTimer
from PyQt6.QtGui import QIcon, QColor
import libonvif as onvif
from cayenue.panels.camera.systemtab import SystemTabSettings
from cayenue.panels.video.modules.settings import VideoModelSettings
from cayenue.enums import ProxyType, StreamState, SnapshotAuth
from loguru import logger

class SessionSignals(QObject):
    finished = pyqtSignal()

class Session(onvif.Session):
    def __init__(self, cp, interface):
        super().__init__()
        self.cp = cp
        self.interface = interface
        self.signals = SessionSignals()
        self.discovered = lambda : self.finish()
        self.getCredential = lambda D : self.cp.getCredential(D)
        self.getData = lambda D : self.cp.getData(D)
        self.infoCallback = lambda msg : self.info(msg)
        self.timer = QTimer()
        self.timer.setSingleShot(True)
        self.timer.timeout.connect(self.timeout)
        self.signals.finished.connect(self.timer.stop)
        self.active = False

    def start(self):
        self.active = True
        self.startDiscover()
        self.timer.start(10000)

    def finish(self):
        self.active = False
        self.signals.finished.emit()
        self.cp.discovered()

    def timeout(self):
        self.cp.discoveryTimeout()

    def info(self, msg):
        try:
            logger.debug(msg)
        except Exception as ex:
            logger.error("session info callback exception: {ex}")

class Camera(QListWidgetItem):
    def __init__(self, onvif_data, mw):
        super().__init__(onvif_data.alias)
        self.mw = mw
        self.icnIdle = QIcon("image:idle_lo.png")
        self.icnOn = QIcon("image:record.png")
        self.icnRecord = QIcon("image:recording_hi.png")
        self.defaultForeground = self.foreground()
        self.filled = False
        self.manual_fill = False
        self.last_msg = ""

        self.assignData(onvif_data)

        module_name = "yolox"
        #if self.mw.videoPanel:
        #module_name = self.mw.videoPanel.panel.module_name    

        self.videoModelSettings = VideoModelSettings(self.mw, self, module_name)
        self.audioModelSettings = None
        self.systemTabSettings = SystemTabSettings(self.mw, self)
        self.manual_recording = False
        self.ordinalKey = f'{self.serial_number()}/Ordinal'
        self.ordinal = self.getOrdinal()
        self.volumeKey = f'{self.serial_number()}/Volume'
        self.volume = self.getVolume()
        self.muteKey = f'{self.serial_number()}/Mute'
        self.mute = self.getMute()
        self.snapshotAuthKey = f'{self.serial_number()}/SnapshotAuth'
        self.snapshotAuth = self.getSnapshotAuth()

    def assignData(self, data):
        self.onvif_data = data
        data.setSetting = self.mw.settings.setValue
        data.getSetting = self.mw.settings.value
        if self.mw.settingsPanel.proxy.proxyType != ProxyType.STAND_ALONE:
            data.getProxyURI = self.mw.getProxyURI

        for profile in data.profiles:
            profile.setSetting = self.mw.settings.setValue
            profile.getSetting = self.mw.settings.value
            if self.mw.settingsPanel.proxy.proxyType != ProxyType.STAND_ALONE:
                profile.getProxyURI = self.mw.getProxyURI
        
        self.profiles = data.profiles

    def uri(self):
        return self.onvif_data.uri()
    
    def serial_number(self):
        return self.onvif_data.serial_number()
    
    def name(self):
        return self.onvif_data.alias
    
    def xaddrs(self):
        return self.onvif_data.xaddrs()

    def hasAudio(self):
        return bool(self.onvif_data.audio_bitrate())
    
    def setOrdinal(self, value):
        self.ordinal = value
        self.mw.settings.setValue(self.ordinalKey, value)

    def getOrdinal(self):
        return int(self.mw.settings.value(self.ordinalKey, -1))
    
    def getMute(self):
        return bool(int(self.mw.settings.value(self.muteKey, 0)))
    
    def setMute(self, state):
        self.mute = bool(state)
        self.mw.settings.setValue(self.muteKey, int(state))

    def getVolume(self):
        return int(self.mw.settings.value(self.volumeKey, 80))
    
    def setVolume(self, volume):
        self.volume = volume
        self.mw.settings.setValue(self.volumeKey, volume)

    def getSnapshotAuth(self):
        return self.mw.settings.value(self.snapshotAuthKey, SnapshotAuth.UNKNOWN)

    def setSnapshotAuth(self, auth):
        self.snapshotAuth = auth
        self.mw.settings.setValue(self.snapshotAuthKey, auth)

    def isRunning(self):
        result = False
        players = self.mw.pm.getStreamPairPlayers(self.uri())
        if len(players):
            result = True
        return result
    
    def isRecording(self):
        result = False
        players = self.mw.pm.getStreamPairPlayers(self.uri())
        for player in players:
            if player.isRecording():
                result = True
        return result
    
    def isAlarming(self):
        result = False
        players = self.mw.pm.getStreamPairPlayers(self.uri())
        for player in players:
            if player.alarm_state:
                result = True
        return result
    
    def isCurrent(self):
        result = False
        for profile in self.profiles:
            camera = self.mw.cameraPanel.getCurrentCamera()
            if camera.getProfile(profile.uri()):
                result = True
        return result

    def editing(self):
        return self.flags() & Qt.ItemFlag.ItemIsEditable
    
    def setIconIdle(self):
        if not self.flags() & Qt.ItemFlag.ItemIsEditable:
            self.setIcon(self.icnIdle)

    def setIconOn(self):
        if not self.flags() & Qt.ItemFlag.ItemIsEditable:
            self.setIcon(self.icnOn)

    def setIconRecord(self):
        if not self.flags() & Qt.ItemFlag.ItemIsEditable:
            self.setIcon(self.icnRecord)
        
    def dimForeground(self):
        self.setForeground(QColor("#808D9E"))

    def restoreForeground(self):
        self.setForeground(self.defaultForeground)

    def isCurrent(self):
        result = False
        current_camera = self.mw.cameraPanel.getCurrentCamera()
        if current_camera:
            if current_camera.serial_number() == self.serial_number():
                result = True
        return result
    
    def getStreamState(self, index):
        result = StreamState.INVALID
        if len(self.profiles) <= index:
            return result
        if profile := self.profiles[index]:
            if player := self.mw.pm.getPlayer(profile.uri()):
                if player.image:
                    result = StreamState.CONNECTED
                else:
                    result = StreamState.CONNECTING
            else:
                result = StreamState.IDLE

            if timer := self.mw.timers.get(profile.uri()):
                if timer.isActive():
                    result = StreamState.CONNECTING
        return result
    
    def profileName(self, uri):
        result = ""
        for profile in self.profiles:
            if profile.uri() == uri:
                result = profile.profile()
        return result
    
    def recordProfileIndex(self):
        return self.systemTabSettings.record_profile

    def displayProfileIndex(self):
        return self.onvif_data.displayProfile
    
    def setDisplayProfile(self, index):
        self.onvif_data.setProfile(index)
        self.mw.settings.setValue(f'{self.serial_number()}/DisplayProfile', index)
        for i, profile in enumerate(self.profiles):
            if i == index:
                profile.setHidden(False)
            else:
                profile.setHidden(True)

    def getDisplayProfileSetting(self):
        return int(self.mw.settings.value(f'{self.serial_number()}/DisplayProfile', self.displayProfileIndex()))

    def getProfile(self, uri):
        result = None
        for profile in self.profiles:
            if  profile.uri() == uri:
                result = profile
                break
        return result
    
    def getRecordProfile(self):
        result = None
        if len(self.profiles) > self.recordProfileIndex():
            result = self.profiles[self.recordProfileIndex()]
        return result
    
    def isRecordProfile(self, uri):
        result = False
        recordProfile = self.getRecordProfile()
        if recordProfile:
            if uri == recordProfile.uri():
                result = True
        return result
    
    def getDisplayProfile(self):
        result = None
        if len(self.profiles) > self.displayProfileIndex():
            result = self.profiles[self.displayProfileIndex()]
        return result
    
    def isDisplayProfile(self, uri):
        result = False
        if displayProfile := self.getDisplayProfile():
            if uri == displayProfile.uri():
                result = True
        return result


    def companionURI(self, uri):
        result = None
        recordProfile = self.getRecordProfile()
        displayProfile = self.getDisplayProfile()
        if recordProfile and displayProfile:
            if uri == recordProfile.uri():
                result = displayProfile.uri()
            if uri == displayProfile.uri():
                result = recordProfile.uri()
        return result

    def syncData(self, data):
        for profile in self.profiles:
            if profile.profile() == data.profile():
                profile.syncData(data)
        data.profiles = self.profiles
        if self.onvif_data.profile() == data.profile():
            self.onvif_data.syncData(data)
            if self.isCurrent():
                self.mw.cameraPanel.signals.fill.emit(data)


#********************************************************************
# Cayenue/cayenue/components/warningbar.py
#
# Copyright (c) 2024  Stephen Rhodes
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#*********************************************************************/

from PyQt6.QtWidgets import QLabel
from PyQt6.QtGui import QPainter, QColorConstants, QLinearGradient, QColor
from PyQt6.QtCore import QRect, QTimer, pyqtSignal, QObject

class WarningBar(QLabel):
    def __init__(self):
        super().__init__()
        self.setMaximumWidth(15)
        self.setStyleSheet("QLabel { border : 1px solid #808D9E; }")
        self.level = 0.0
        self.inverted = False

    def paintEvent(self, event):
        marker = max(int(self.height()*(1-self.level)-2), 0)
        if self.inverted:
            marker = min(int(self.height() * self.level), self.height()-2)
        painter = QPainter(self)
        gradient = QLinearGradient(0,0,0,100)
        gradient.setColorAt(0.0, QColorConstants.Red)
        gradient.setColorAt(0.5, QColorConstants.DarkYellow)
        gradient.setColorAt(1.0, QColorConstants.DarkGreen)
        painter.fillRect(QRect(1, 1, 13, self.height()-2), gradient)
        painter.fillRect(QRect(1, 1, 13, marker), QColor("#3B3B3B"))

    def setLevel(self, level):
        self.level = level
        self.update()

class IndicatorSignals(QObject):
    start = pyqtSignal()

class Indicator(QLabel):
    def __init__(self, mw):
        super().__init__()
        self.mw = mw
        self.setMaximumWidth(15)
        self.setMaximumHeight(10)
        self.setStyleSheet("QLabel { border : 1px solid #808D9E; }")
        self.timer = QTimer()
        self.timer.setInterval(self.mw.settingsPanel.alarm.spnLagTime.value() * 1000)
        self.timer.setSingleShot(True)
        self.timer.timeout.connect(self.timeout)
        self.signals = IndicatorSignals()
        self.signals.start.connect(self.timer.start)
        self.state = 0

    def paintEvent(self, event):
        color = QColor("#3B3B3B")
        if self.state:
            color = QColorConstants.Red
        painter = QPainter(self)
        painter.fillRect(QRect(1, 1, 13, self.height()-2), color)

    def setState(self, state):
        self.state = int(state)
        if state:
            self.signals.start.emit()
        self.update()

    def getState(self):
        return self.state

    def timeout(self):
        self.setState(0)




from .networktab import NetworkTab
from .videotab import VideoTab
from .imagetab import ImageTab
from .systemtab import SystemTab
from .logindialog import LoginDialog
from .ptztab import PTZTab
from .datastructures import Session, Camera
from .cameralist import CameraList
from .snapshot import Snapshot